#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

from __future__ import division

__all__ = ['Expression', '_EqualityExpression', '_InequalityExpression',
           '_IdentityExpression' ,
           'generate_expression', 'generate_relational_expression']

from plugin import *
from pyutilib.component.core import *
from numvalue import *
from param import _ParamBase
from var import _VarBase

import sys
import copy
import logging
import StringIO

logger = logging.getLogger('coopr.pyomo')

def clone_expression(exp):
    if exp.is_expression():
        return copy.copy(exp)
    else:
        return exp

def chainedInequalityErrorMessage(msg=None):
    if msg is None:
        msg = "Nonconstant relational expression used in an "\
              "unexpected Boolean context."
    buf = StringIO.StringIO()
    generate_relational_expression.chainedInequality.pprint(buf)
    # We are about to raise an exception, so it's OK to reset chainedInequality
    generate_relational_expression.chainedInequality = None
    msg += """
The inequality expression:
    %s
contains non-constant terms (variables) appearing in a Boolean context, e.g.:
    if expression <= 5:
This is generally invalid.  If you want to obtain the Boolean value of
the expression based on the current variable values, explicitly evaluate
the expression, e.g.:
    if value(expression) <= 5:
or
    if value(expression <= 5):
""" % ( buf.getvalue().strip(), )
    return msg
    

class Expression(NumericValue):
    """An object that defines a mathematical expression that can be evaluated"""

    __slots__ = ('_args', )

    def __init__(self, name, nargs, args):
        """Construct an expression with an operation and a set of arguments"""
        NumericValue.__init__(self, name, None, None, False)
        self._args=args
        if nargs and nargs != len(args):
            raise ValueError, "%s() takes exactly %d arguments (%d given)" % \
                ( name, nargs, len(args) )

    def __getstate__(self):
       result = NumericValue.__getstate__(self)
       for i in Expression.__slots__:
          result[i] = getattr(self, i)
       return result
   
    def pprint(self, ostream=None, nested=True, eol_flag=True):
        """Print this expression"""
        if ostream is None:
           ostream = sys.stdout
        if nested:
           if self.name is None:
              print >>ostream, "Unnamed-Expression" + "(",
           else:
              print >>ostream, self.name + "(",
           first=True
           for arg in self._args:
             if first==False:
                print >>ostream, ",",
             if arg.is_expression():
                arg.pprint(ostream=ostream, nested=nested, eol_flag=False)
             else:
                print >>ostream, str(arg),
             first=False
           if eol_flag==True:
              print >>ostream, ")"
           else:
              print >>ostream, ")",

    def clone(self):
        return copy.copy(self)

    def __copy__(self):
        """Clone this object using the specified arguments"""
        raise NotImplementedError, "Derived expression (%s) failed to "\
            "implement __copy__()" % ( str(self.__class__), )

    def simplify(self, model): #pragma:nocover
        print """
WARNING: Expression.simplify() has been deprecated and removed from
     Pyomo Expressions.  Please remove references to simplify() from your
     code.
"""
        return self

    #
    # this method contrast with the fixed_value() method.
    # the fixed_value() method returns true iff the value is
    # an atomic constant. 
    # this method returns true iff all composite arguments
    # in this sum expression are constant, i.e., numeric 
    # constants or parametrs. the parameter values can of 
    # course change over time, but at any point in time, 
    # they are constant. hence, the name.
    #
    def is_constant(self):
        for arg in self._args:
            if not arg.is_constant():
                return False
        return True

    def is_expression(self):
        return True

    def polynomial_degree(self):
        return None

    def __nonzero__(self):
        return bool(self())

    def __call__(self, exception=True):
        """Evaluate the expression"""
        try:
            return self._apply_operation(self._evaluate_arglist(self._args))
        except (ValueError, TypeError):
            if exception:
                raise
            return None

    def _evaluate_arglist(self, arglist):
        for arg in arglist:
            try:
                yield value(arg)
            except Exception, e:
                buffer = StringIO.StringIO()
                self.pprint(buffer)
                logger.error("evaluating expression: %s\n    (expression: %s)",
                             str(e), buffer.getvalue())
                raise

    def _apply_operation(self, values):
        """Method that can be overwritten to define the operation in
        this expression"""
        raise NotImplementedError, "Derived expression (%s) failed to "\
            "implement _apply_operation()" % ( str(self.__class__), )
        
    def __str__(self):
        return self.name


class _IntrinsicFunctionExpression(Expression):
    __slots__ = ('_operator', )

    def __init__(self, name, nargs, args, operator):
        """Construct an expression with an operation and a set of arguments"""
        Expression.__init__(self, name, nargs, args)
        self._operator = operator

    def __getstate__(self):
       result = Expression.__getstate__(self)
       for i in _IntrinsicFunctionExpression.__slots__:
          result[i] = getattr(self, i)
       return result

    def __copy__(self):
        return self.__class__( self.name, 
                               None, 
                               tuple(clone_expression(x) for x in self._args),
                               self._operator )

    def _apply_operation(self, values):
        return self._operator(*tuple(values))


# Should this actually be a special class, or just an instance of
# _IntrinsicFunctionExpression (like sin, cos, etc)?
class _AbsExpression(_IntrinsicFunctionExpression):

    __slots__ = ()

    def __init__(self, args):
        _IntrinsicFunctionExpression.__init__(self, 'abs', 1, args, abs)

    def __getstate__(self):
       return _IntrinsicFunctionExpression.__getstate__(self)

    def __copy__(self):
        return self.__class__( tuple(clone_expression(x) for x in self._args) )

# Should this actually be a special class, or just an instance of
# _IntrinsicFunctionExpression (like sin, cos, etc)?
class _PowExpression(_IntrinsicFunctionExpression):

    __slots__ = ()

    def __init__(self, args):
        _IntrinsicFunctionExpression.__init__(self, 'pow', 2, args, pow)

    def __getstate__(self):
       return _IntrinsicFunctionExpression.__getstate__(self)
 
    def __copy__(self):
        return self.__class__( tuple(clone_expression(x) for x in self._args) )

    def polynomial_degree(self):
        # Right now, all _PowExpressions are considered to be
        # non-polynomial, even though there is a special case where
        # _args[0] is polynomial and _args[1] is constant and a
        # non-negative integer.  I am still choosing to return
        # "nonpolynomial" because most consumers (like expression
        # compilers) do not recognize that pow could actually be
        # polynomial.
        return None


class _LinearExpression(Expression):

    __slots__ = ()

    def __init__(self, name, nargs, args):
        """Constructor"""
        Expression.__init__(self, name, nargs, args)

    def __getstate__(self):
       return Expression.__getstate__(self)

    def polynomial_degree(self):
        # NB: We can't use max() here because None (non-polynomial)
        # overrides a numeric value (and max() just ignores it)
        degree = 0
        for x in self._args:
            x_degree = x.polynomial_degree()
            if x_degree is None:
                return None
            degree = max(degree, x_degree)
        return degree


class _InequalityExpression(_LinearExpression):
    """An object that defines a series of less-than or
    less-than-or-equal expressions"""

    __slots__ = ('_strict', '_cloned_from')

    def __init__(self, args, strict, cloned_from):
        """Constructor"""
        _LinearExpression.__init__(self, 'ineq', None, args)
        self._strict = strict
        self._cloned_from = cloned_from
        if len(self._args)-1 != len(self._strict):
            raise ValueError, "_InequalityExpression(): len(args)-1 != " \
                  "len(strict)(%d and %d given)" % (len(args), len(strict))

    def __getstate__(self):
       result = _LinearExpression.__getstate__(self)
       for i in _InequalityExpression.__slots__:
          result[i] = getattr(self, i)
       return result

    def __nonzero__(self):
        if generate_relational_expression.chainedInequality is not None:
            raise TypeError, chainedInequalityErrorMessage()
        if not self.is_constant():
            generate_relational_expression.chainedInequality = self
            return True

        return bool(self())

    def is_relational(self):
        return True

    def __copy__(self):
        return self.__class__( [clone_expression(x) for x in self._args],
                               copy.copy(self._strict),
                               copy.copy(self._cloned_from) )

    def _apply_operation(self, values):
        """Method that defines the less-than-or-equal operation"""
        arg1 = values.next()
        for i in xrange(len(self._strict)):
            arg2 = values.next()
            if self._strict[i]:
                if not (arg1 < arg2):
                    return False
            else:
                if not (arg1 <= arg2):
                    return False
            arg1 = arg2
        return True

    def pprint(self, ostream=None, nested=True, eol_flag=True):
        """Print this expression"""
        if ostream is None:
           ostream = sys.stdout
        if nested:
           first=True
           for i in xrange(len(self._strict)):
               arg = self._args[i]
               if arg.is_expression():
                   arg.pprint(ostream=ostream, nested=nested, eol_flag=False)
               else:
                   print >>ostream, str(arg),
               if self._strict[i]:
                   print >> ostream, " < ",
               else:
                   print >> ostream, " <= ",
           arg = self._args[-1]
           if arg.is_expression():
               arg.pprint(ostream=ostream, nested=nested, eol_flag=False)
           else:
               print >>ostream, str(arg),
           if eol_flag==True:
              print >>ostream, ""


class _EqualityExpression(_LinearExpression):
    """An object that defines a equal-to expression"""

    __slots__ = ()

    def __init__(self, args):
        """Constructor"""
        _LinearExpression.__init__(self, 'eq', 2, args)

    def __getstate__(self):
       return _LinearExpression.__getstate__(self)

    def __copy__(self):
        return self.__class__( tuple(clone_expression(x) for x in self._args) )

    def __nonzero__(self):
        if generate_relational_expression.chainedInequality is not None:
            raise TypeError, chainedInequalityErrorMessage()
        return bool(self())

    def is_relational(self):
        return True

    def _apply_operation(self, values):
        """Method that defines the equal-to operation"""
        return values.next() == values.next()

    def pprint(self, ostream=None, nested=True, eol_flag=True):
        """Print this expression"""
        if ostream is None:
           ostream = sys.stdout
        if nested:
            first = True
            for arg in self._args:
                if not first:
                    print >> ostream, " == ",
                first = False
                if arg.is_expression():
                    arg.pprint(ostream=ostream, nested=nested, eol_flag=False)
                else:
                    print >>ostream, str(arg),
            if eol_flag==True:
                print >>ostream, ""

class _ProductExpression(Expression):
    """An object that defines a product expression"""

    __slots__ = ('_denominator','_numerator','coef')

    def __init__(self):
        """Constructor"""
        Expression.__init__(self,'prod',None, None)
        self._denominator = []
        self._numerator = []
        self.coef = 1

    def __getstate__(self):
       result = Expression.__getstate__(self)
       for i in _ProductExpression.__slots__:
          result[i] = getattr(self, i)
       return result        

    def is_constant(self):
        for arg in self._numerator:
            if not arg.is_constant():
                return False
        for arg in self._denominator:
            if not arg.is_constant():
                return False
        return True

    def polynomial_degree(self):
        for x in self._denominator:
            if x.polynomial_degree() != 0:
                return None
        try:
            return sum(x.polynomial_degree() for x in self._numerator)
        except TypeError:
            return None

    def invert(self):
        tmp = self._denominator
        self._denominator = self._numerator
        self._numerator = tmp
        self.coef = 1.0/self.coef
            
    def pprint(self, ostream=None, nested=True, eol_flag=True):
        """Print this expression"""
        if ostream is None:
           ostream = sys.stdout
        if nested:
           print >>ostream, self.name + "( num=(",
           first=True
           if self.coef != 1:
                print >>ostream, str(self.coef),
                first=False
           for arg in self._numerator:
             if first==False:
                print >>ostream, ",",
             if arg.is_expression():
                arg.pprint(ostream=ostream, nested=nested, eol_flag=False)
             else:
                print >>ostream, str(arg),
             first=False
           if first is True:
              print >>ostream, 1,
           print >>ostream, ")",
           if len(self._denominator) > 0:
              print >>ostream, ", denom=(",
              first=True
              for arg in self._denominator:
                if first==False:
                   print >>ostream, ",",
                if arg.is_expression():
                   arg.pprint(ostream=ostream, nested=nested, eol_flag=False)
                else:
                   print >>ostream, str(arg),
                first=False
              print >>ostream, ")",
           print >>ostream, ")",
           if eol_flag==True:
              print >>ostream, ""

    def __copy__(self):
        """Clone this object using the specified arguments"""
        tmp = self.__class__()
        tmp.name = self.name
        tmp._numerator = [clone_expression(x) for x in self._numerator]
        tmp._denominator = [clone_expression(x) for x in self._denominator]
        tmp.coef = self.coef
        return tmp

    def __call__(self, exception=True):
        """Evaluate the expression"""
        try:
            ans = self.coef
            for n in self._evaluate_arglist(self._numerator):
                ans *= n
            for n in self._evaluate_arglist(self._denominator):
                ans /= n 
            return ans
        except (TypeError, ValueError):
            if exception:
                raise
            return None


class _IdentityExpression(_LinearExpression):
    """An object that defines a identity expression"""

    __slots__ = ()

    def __init__(self, args):
        """Constructor"""
        arg_type = type(args)
        if arg_type is list or arg_type is tuple:
           Expression.__init__(self, 'identity', 1, tuple(args))
        else:
           Expression.__init__(self, 'identity', 1, (args,))

    def __getstate__(self):
       return _LinearExpression.__getstate__(self)

    def __copy__(self):
        return self.__class__( tuple(clone_expression(x) for x in self._args) )

    def _apply_operation(self, values):
        """Method that defines the identity operation"""
        return values.next()


class _SumExpression(_LinearExpression):
    """An object that defines a weighted summation of expressions"""

    __slots__ = ('_coef','_const')

    def __init__(self):
        """Constructor"""
        _LinearExpression.__init__(self, 'sum', None, [])
        self._coef = []
        self._const = 0

    def __getstate__(self):
       result = _LinearExpression.__getstate__(self)
       for i in _SumExpression.__slots__:
          result[i] = getattr(self, i)
       return result        

    def __copy__(self):
        """Clone this object using the specified arguments"""
        tmp = self.__class__()
        tmp.name = self.name
        tmp._args = [clone_expression(x) for x in self._args]
        tmp._coef = copy.copy(self._coef)
        tmp._const = self._const
        return tmp

    def scale(self, val):
        for i in xrange(len(self._coef)):
            self._coef[i] *= val
        self._const *= val

    def negate(self):
        self.scale(-1)

    def pprint(self, ostream=None, nested=True, eol_flag=True):
        """Print this expression"""
        if ostream is None:
           ostream = sys.stdout
        if nested:
           print >>ostream, self.name + "(",
           first=True
           if self._const != 0:
                print >>ostream, str(self._const),
                first=False
           for i, arg in enumerate(self._args):
             if first==False:
                print >>ostream, ",",
             if self._coef[i] != 1:
                print >>ostream, str(self._coef[i])+" * ",
             if arg.is_expression():
                arg.pprint(ostream=ostream, nested=nested, eol_flag=False)
             else:
                print >>ostream, str(arg),
             first=False
           print >>ostream, ")",
           if eol_flag==True:
              print >>ostream, ""

    def _apply_operation(self, values):
        """Evaluate the expression"""
        return sum(c*values.next() for c in self._coef) + self._const



def generate_expression(*_args):
    def _clone_if_needed(obj, target):
        count = sys.getrefcount(obj) - \
                generate_expression.UNREFERENCED_EXPR_COUNT
        if generate_expression.clone_if_needed_callback:
            generate_expression.clone_if_needed_callback(count)
        if count == target:
            return obj
        elif count > target:
            generate_expression.clone_counter += 1
            return obj.clone()
        else: #pragma:nocover
            raise RuntimeError, "Expression entered generate_expression() " \
                "with too few references (%s<0); this is indicative of a " \
                "SERIOUS ERROR in the expression reuse detection scheme." \
                % ( count, )

    etype = _args[0]
    _self = _args[1]

    # In-place operators should only clone `self` if someone else (other
    # than the self) holds a reference to it.  This also enforces that
    # there MUST be an external `self` reference for in-place operators.
    if etype[0] == 'i':
        etype = etype[1:]
        if _self.is_expression():
            _self = _clone_if_needed(_self, 1)
    else:
        if _self.is_expression():
            _self = _clone_if_needed(_self, 0)

    if _self.is_indexed():
        raise TypeError, "Argument for expression '%s' is an indexed "\
              "numeric value specified without an index: %s\n    Is "\
              "variable or parameter '%s' defined over an index that "\
              "you did not specify?" % (etype, _self.name, _self.name)
        
    #
    # First, handle the special cases of unary opertors
    #
    if etype == 'neg':
        if type(_self) is _SumExpression:
            _self.negate()
            return _self
        elif type(_self) is _ProductExpression:
            _self.coef *= -1
            return _self
        else:
            etype = 'rmul'
            other = NumericConstant(None, None, -1)
    elif etype == 'abs':
        return _AbsExpression([_self])
    else:
        other = as_numeric(_args[2])
        if other.is_expression():
            other = _clone_if_needed(other, 0)
        elif other.is_indexed():
            raise TypeError, "Argument for expression '%s' is an indexed "\
                  "numeric value specified without an index: %s\n    Is "\
                  "variable or parameter '%s' defined over an index that "\
                  "you did not specify?" % (etype, other.name, other.name)

    #
    # Binary operators can either be "normal" or "reversed"; reverse all
    # "reversed" opertors
    #
    if etype[0] == 'r':
        #
        # This may seem obvious, but if we are performing an
        # "R"-operation (i.e. reverse operation), then simply reverse
        # self and other.  This is legitimate as we are generating a
        # completely new expression here, and the _clone_if_needed logic
        # above will make sure that we don't accidentally clobber
        # someone else's expression (fragment).
        #
        tmp = _self
        _self = other
        other = tmp
        etype = etype[1:]

    #
    # Some binary operators are special cases of other operators
    #
    multiplier = 1
    if etype == 'sub':
        multiplier = -1
        etype = 'add'

    #
    # Now, handle all binary operators
    #
    if etype == 'add':
        #
        # self + other 
        #
        self_type = type(_self)
        other_type = type(other)
        if self_type is _SumExpression:
            if other_type is NumericConstant:
                _self._const += multiplier * other()
            elif other_type is _ProductExpression \
                     and len(other._numerator) == 1 \
                     and not other._denominator:
                _self._args.append(other._numerator[0])
                _self._coef.append(multiplier*other.coef)
            elif other_type is _SumExpression:
                if multiplier < 0:
                    other.negate()
                _self._args.extend(other._args)
                _self._coef.extend(other._coef)
                _self._const += other._const
            else:
                _self._args.append(other)
                _self._coef.append(multiplier)
            ans = _self
        else:
            if other_type is _SumExpression:
                if multiplier < 0:
                    other.negate()
                if self_type is NumericConstant:
                    other._const += _self()
                elif self_type is _ProductExpression and \
                         len(_self._numerator) == 1 and \
                         not _self._denominator:
                    other._args.insert(0,_self._numerator[0])
                    other._coef.insert(0,_self.coef)
                else:
                    other._args.insert(0,_self)
                    other._coef.insert(0,1)
                ans = other
            else:
                ans = _SumExpression()
                if self_type is NumericConstant:
                    ans._const += _self()
                elif self_type is _ProductExpression and \
                         len(_self._numerator) == 1 and \
                         not _self._denominator:
                    ans._args.append(_self._numerator[0])
                    ans._coef.append( _self.coef)
                else:
                    ans._args.append(_self)
                    ans._coef.append(1)
                if other_type is NumericConstant:
                    ans._const += multiplier * other()
                elif other_type is _ProductExpression and \
                         len(other._numerator) == 1 and \
                         not other._denominator:
                    ans._args.append(other._numerator[0])
                    ans._coef.append( multiplier*other.coef)
                else:
                    ans._args.append(other)
                    ans._coef.append( multiplier)

        # Special cases for simplifying expressions
        l = len(ans._args)
        if l == 0:
            return NumericConstant(None, None, ans._const)
        elif l == 1 and ans._const == 0 and ans._coef[0] == 1:
            return ans._args[0]
        return ans

    elif etype == 'mul':
        #
        # self * other
        #
        self_type = type(_self)
        other_type = type(other)
        if self_type is _ProductExpression:
            if other_type is NumericConstant:
                _self.coef *= other()
            elif other_type is _ProductExpression:
                _self._numerator.extend(other._numerator)
                _self._denominator.extend(other._denominator)
                _self.coef *= other.coef
            else:
                _self._numerator.append(other)
            ans = _self
        else:
            if other_type is _ProductExpression:
                if self_type is NumericConstant:
                    other.coef *= _self()
                else:
                    other._numerator.insert(0,_self)
                ans = other
            else:
                ans = _ProductExpression()
                if self_type is NumericConstant:
                    ans.coef *= _self()
                else:
                    ans._numerator.append(_self)
                if other_type is NumericConstant:
                    ans.coef *= other()
                else:
                    ans._numerator.append(other)

        # Special cases for simplifying expressions
        if ans.coef == 0:
            return NumericConstant(None,None,0)
        if not ans._denominator:
            l = len(ans._numerator)
            if l == 0:
                return NumericConstant(None,None,ans.coef)
            elif l == 1 and ans.coef == 1:
                return ans._numerator[0]
        return ans

    elif etype == 'div':
        #
        # self / other
        #
        self_type = type(_self)
        other_type = type(other)
        if self_type is _ProductExpression:
            if other_type is NumericConstant:
                _self.coef /= other()
            elif other_type is _ProductExpression:
                _self._numerator.extend(other._denominator)
                _self._denominator.extend(other._numerator)
                _self.coef /= other.coef
            else:
                _self._denominator.append(other)
            ans = _self
        else:
            if other_type is _ProductExpression:
                other.invert()
                if self_type is NumericConstant:
                    other.coef *= _self()
                else:
                    other._numerator.insert(0,_self)
                ans = other
            else:
                ans = _ProductExpression()
                if self_type is NumericConstant:
                    ans.coef *= _self()
                else:
                    ans._numerator.append(_self)
                if other_type is NumericConstant:
                    ans.coef /= other()
                else:
                    ans._denominator.append(other)

        # Special cases for simplifying expressions
        if ans.coef == 0:
            return NumericConstant(None,None,0)
        if not ans._denominator:
            l = len(ans._numerator)
            if l == 0:
                return NumericConstant(None,None,ans.coef)
            elif l == 1 and ans.coef == 1:
                return ans._numerator[0]
        return ans

    elif etype == 'pow':
        #
        # self ** other
        #
        return _PowExpression((_self, other))

    else:
        raise RuntimeError, "Unknown expression type '%s'" % etype

##
## "static" variables within the generate_expression function
##

# [testing] clone_check_callback allows test functions to intercept the
# call to _clone_if_needed and see the returned value from
# sys.getrefcount.
generate_expression.clone_if_needed_callback = None

# [debugging] clone_counter is a count of the number of calls to
# expr.clone() made during expression generation.
generate_expression.clone_counter = 0

# [configuration] UNREFERENCED_EXPR_COUNT is a "magic number" that
# indicates the stack depth between "normal" modeling and
# _clone_if_needed().  If an expression enters _clone_if_needed() with
# UNREFERENCED_EXPR_COUNT references, then there are no other variables
# that hold a reference to the expression and cloning is not necessary.
# If there are more references than UNREFERENCED_EXPR_COUNT, then we
# must clone the expression before operating on it.  It should be an
# error to hit _clone_if_needed() with fewer than
# UNREFERENCED_EXPR_COUNT references.
generate_expression.UNREFERENCED_EXPR_COUNT = 10



def generate_relational_expression(etype, lhs, rhs):
    def _clone_if_needed(obj):
        count = sys.getrefcount(obj) - \
                generate_relational_expression.UNREFERENCED_EXPR_COUNT
        if generate_relational_expression.clone_if_needed_callback:
            generate_relational_expression.clone_if_needed_callback(count)
        if count == 0 or \
                count == generate_relational_expression.CHAINED_EXPR_COUNT:
            return obj
        elif count > 0:
            generate_relational_expression.clone_counter += 1
            return obj.clone()
        else:
            raise RuntimeError, "Expression entered " \
                "generate_relational_expression() " \
                "with too few references (%s<0); this is indicative of a " \
                "SERIOUS ERROR in the expression reuse detection scheme." \
                % ( count, )

    cloned_from = (id(lhs), id(rhs))
    if generate_relational_expression.chainedInequality is not None:
        prevExpr = generate_relational_expression.chainedInequality
        for arg in prevExpr._cloned_from:
            if arg == id(lhs):
                lhs = prevExpr
                prevExpr = None
                cloned_from = lhs._cloned_from + (cloned_from[1],)
                break
            if arg == id(rhs):
                rhs = prevExpr
                prevExpr = None
                cloned_from = (cloned_from[0],) + rhs._cloned_from
                break
        if prevExpr is not None:
            raise TypeError, chainedInequalityErrorMessage()
        generate_relational_expression.chainedInequality = None

    if lhs.is_expression():
        lhs = _clone_if_needed(lhs)
    elif lhs.is_indexed():
        raise TypeError, "Argument for expression '%s' is an indexed "\
              "numeric value specified without an index: %s\n    Is "\
              "variable or parameter '%s' defined over an index that "\
              "you did not specify?" % (etype, lhs.name, lhs.name)

    if rhs.is_expression():
        rhs = _clone_if_needed(rhs)
    elif rhs.is_indexed():
        raise TypeError, "Argument for expression '%s' is an indexed "\
              "numeric value specified without an index: %s\n    Is "\
              "variable or parameter '%s' defined over an index that "\
              "you did not specify?" % (etype, rhs.name, rhs.name)

    if etype == '==':
        if lhs.is_relational() or rhs.is_relational():
            buf = StringIO.StringIO()
            if lhs.is_relational():
                lhs.pprint(buf)
            else:
                rhs.pprint(buf)
            raise TypeError, "Cannot create an EqualityExpression where "\
                  "one of the sub-expressions is a relational expression:\n"\
                  "    " + buf.getvalue().strip()
        return _EqualityExpression((lhs,rhs))
    else:
        if etype == '<=':
            strict = False
        elif etype == '<':
            strict = True
        else:
            raise ValueError, "Unknown relational expression type '%s'" % etype
        if lhs.is_relational():
            if lhs.__class__ is _InequalityExpression:
                if rhs.is_relational():
                    raise TypeError, "Cannot create an InequalityExpression "\
                          "where both sub-expressions are also relational "\
                          "expressions (we support no more than 3 terms "\
                          "in an inequality expression)."
                if len(lhs._args) > 2:
                    raise ValueError, "Cannot create an InequalityExpression "\
                          "with more than 3 terms."
                lhs._args.append(rhs)
                lhs._strict.append(strict)
                lhs._cloned_from = cloned_from
                return lhs
            else:
                buf = StringIO.StringIO()
                lhs.pprint(buf)
                raise TypeError, "Cannot create an InequalityExpression "\
                      "where one of the sub-expressions is an equality "\
                      "expression:\n    " + buf.getvalue().strip()
        elif rhs.is_relational():
            if rhs.__class__ is _InequalityExpression:
                if len(rhs._args) > 2:
                    raise ValueError, "Cannot create an InequalityExpression "\
                          "with more than 3 terms."
                rhs._args.insert(0, lhs)
                rhs._strict.insert(0, strict)
                rhs._cloned_from = cloned_from
                return rhs
            else:
                buf = StringIO.StringIO()
                rhs.pprint(buf)
                raise TypeError, "Cannot create an InequalityExpression "\
                      "where one of the sub-expressions is an equality "\
                      "expression:\n    " + buf.getvalue().strip()
        else:
            return _InequalityExpression([lhs,rhs], [strict], cloned_from)


##
## "static" variables within the generate_relational_expression function
##

# [functionality] chainedInequality allows us to generate symbolic
# expressions of the type "a < b < c".  This provides a buffer to hold
# the first inequality so the second inequality can access it later.
generate_relational_expression.chainedInequality = None

# [testing] clone_check_callback allows test functions to intercept the
# call to _clone_if_needed and see the returned value from
# sys.getrefcount.
generate_relational_expression.clone_if_needed_callback = None

# [debugging] clone_counter is a count of the number of calls to
# expr.clone() made during expression generation.
generate_relational_expression.clone_counter = 0

# [configuration] UNREFERENCED_EXPR_COUNT is a "magic number" that
# indicates the stack depth between "normal" modeling and
# _clone_if_needed().  If an expression enters _clone_if_needed() with
# UNREFERENCED_EXPR_COUNT references, then there are no other variables
# that hold a reference to the expression and cloning is not necessary.
# If there are more references than UNREFERENCED_EXPR_COUNT, then we
# must clone the expression before operating on it.  It should be an
# error to hit _clone_if_needed() with fewer than
# UNREFERENCED_EXPR_COUNT references.
generate_relational_expression.UNREFERENCED_EXPR_COUNT = 9

# [configuration] CHAINED_EXPR_COUNT is a "magic number" that indicates
# the relative stack depth for compound inequality expressions of the
# type "a < b < c".  Since Python creates the first inequality as
# temporary, and releases it after evaluating __nonzero__(), that
# expression's refcount will be CHAINED_EXPR_COUNT less than the normal
# UNREFERENCED_EXPR_COUNT.
generate_relational_expression.CHAINED_EXPR_COUNT = -5


if False: #pragma:nocover
    ExpressionRegistration('<', _InequalityExpression)
    ExpressionRegistration('lt', _InequalityExpression)
    ExpressionRegistration('>', _InequalityExpression, True)
    ExpressionRegistration('gt', _InequalityExpression, True)
    ExpressionRegistration('<=', _InequalityExpression)
    ExpressionRegistration('lte', _InequalityExpression)
    ExpressionRegistration('>=', _InequalityExpression, True)
    ExpressionRegistration('gte', _InequalityExpression, True)
    ExpressionRegistration('=', _EqualityExpression)
    ExpressionRegistration('eq', _EqualityExpression)

if False: #pragma:nocover
    ExpressionRegistration('+', _SumExpression)
    ExpressionRegistration('sum', _SumExpression)
    ExpressionRegistration('*', _ProductExpression)
    ExpressionRegistration('prod', _ProductExpression)
    #ExpressionRegistration('-', _MinusExpression)
    #ExpressionRegistration('minus', _MinusExpression)
    #ExpressionRegistration('/', _DivisionExpression)
    #ExpressionRegistration('divide', _DivisionExpression)
    #ExpressionRegistration('-', _NegateExpression)
    #ExpressionRegistration('negate', _NegateExpression)
    ExpressionRegistration('abs', _AbsExpression)
    ExpressionRegistration('pow', _PowExpression)




#
# Get the directory where this script is defined, and where the baseline
# files are located.
#
import os
import sys
import string
from os.path import abspath, dirname
sys.path.insert(0, dirname(dirname(abspath(__file__)))+"/../..")

this_test_directory = dirname(abspath(__file__))+os.sep

benders_example_dir = dirname(dirname(dirname(dirname(dirname(abspath(__file__))))))+os.sep+"examples"+os.sep+"pyomo"+os.sep+"benders"+os.sep

coopr_bin_dir = dirname(dirname(dirname(dirname(dirname(dirname(dirname(abspath(__file__))))))))+os.sep+"bin"+os.sep

#
# Import the testing packages
#
import pyutilib.misc
import pyutilib.th as unittest
import pyutilib.subprocess
import coopr.plugins.mip

cplex = None
cplex_available = False
try:
    cplex = coopr.plugins.mip.CPLEX(keepFiles=True)
    cplex_available = (not cplex.executable() is None) and cplex.available(False)
except pyutilib.common.ApplicationError:
    cplex_available=False

gurobi = None
gurobi_available = False
try:
    gurobi = coopr.plugins.mip.GUROBI(keepFiles=True)
    gurobi_available = (not gurobi.executable() is None) and gurobi.available(False)
except pyutilib.common.ApplicationError:
    gurobi_available=False

glpk = None
glpk_available = False
try:
    glpk = coopr.plugins.mip.GLPK(keepFiles=True)
    glpk_available = (not glpk.executable() is None) and glpk.available(False)
except pyutilib.common.ApplicationError:
    glpk_available=False

class TestBenders(unittest.TestCase):

    def cleanup(self):

        # IMPT: This step is key, as Python keys off the name of the module, not the location.
        #       So, different reference models in different directories won't be detected.
        #       If you don't do this, the symptom is a model that doesn't have the attributes
        #       that the data file expects.
        if "ReferenceModel" in sys.modules:
           del sys.modules["ReferenceModel"]

    @unittest.skipIf(not cplex_available, "The 'cplex' executable is not available")
    def test_benders_cplex(self):
        pyutilib.misc.setup_redirect(this_test_directory+"benders_cplex.out")
        os.chdir(benders_example_dir)
        os.system("lbin python "+benders_example_dir+"runbenders")
        os.chdir(this_test_directory)        
        pyutilib.misc.reset_redirect()
        self.cleanup()        
        self.failUnlessFileEqualsBaseline(this_test_directory+"benders_cplex.out",this_test_directory+"benders_cplex.baseline")

TestBenders = unittest.category('smoke')(TestBenders)

if __name__ == "__main__":
    unittest.main()



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

import gc
import logging
import os
import sys
import textwrap
import traceback
import types
import time

try:
    import cProfile as profile
except ImportError:
    import profile
try:
    import pstats
    pstats_available=True
except ImportError:
    pstats_available=False

try:
    import IPython
    IPython_available=True
    from IPython.Shell import IPShellEmbed
    ipshell = IPShellEmbed([''],
                banner = '\n# Dropping into Python interpreter',
                exit_msg = '\n# Leaving Interpreter, back to Pyomo\n')
except ImportError:
    IPython_available=False

from pyutilib.misc import Options
try:
   from pympler.muppy import muppy
   from pympler.muppy import summary
   from pympler.muppy import tracker
   from pympler.asizeof import *
   pympler_available = True
except ImportError:
   pympler_available = False
memory_tracker = None
memory_data = Options()

from coopr.pyomo import *
from coopr.opt import ProblemFormat
from coopr.opt.base import SolverFactory
from coopr.opt.parallel import SolverManagerFactory
import pyutilib.misc
from pyutilib.component.core import ExtensionPoint, Plugin, implements
from pyutilib.services import TempfileManager

from coopr.pyomo.expr.linear_repn import linearize_model_expressions

filter_excepthook=False
modelapi = {    'pyomo_create_model':IPyomoScriptCreateModel,
                'pyomo_create_modeldata':IPyomoScriptCreateModelData,
                'pyomo_print_model':IPyomoScriptPrintModel,
                'pyomo_modify_instance':IPyomoScriptModifyInstance,
                'pyomo_print_instance':IPyomoScriptPrintInstance,
                'pyomo_save_instance':IPyomoScriptSaveInstance,
                'pyomo_print_results':IPyomoScriptPrintResults,
                'pyomo_save_results':IPyomoScriptSaveResults,
                'pyomo_postprocess':IPyomoScriptPostprocess}


logger = logging.getLogger('coopr.pyomo')
start_time = 0.0

#
# Print information about modeling components supported by Pyomo
#
def print_components(options):
    print ""
    print "----------------------------------------------------------------"
    print "Pyomo Model Components:"
    print "----------------------------------------------------------------"
    components = pyomo.model_components()
    index = pyutilib.misc.sort_index(components)
    for i in index:
        print ""
        print " "+components[i][0]
        for line in textwrap.wrap(components[i][1], 59):
            print "    "+line
    print ""
    print "----------------------------------------------------------------"
    print "Pyomo Virtual Sets:"
    print "----------------------------------------------------------------"
    pyomo_sets = pyomo.predefined_sets()
    index = pyutilib.misc.sort_index(pyomo_sets)
    for i in index:
        print ""
        print " "+pyomo_sets[i][0]
        print "    "+pyomo_sets[i][1]

#
# Print information about the solvers that are available
#
def print_solver_help(options):
    print "The following solvers are are currently supported:"
    solver_list = SolverFactory.services()
    solver_list = sorted( filter(lambda x: '_' != x[0], solver_list) )
    n = max(map(len, solver_list))
    wrapper = textwrap.TextWrapper(subsequent_indent=' '*(n+9))
    for s in solver_list:
        format = '    %-'+str(n)+'s  %s'
        print wrapper.fill(format % (s , SolverFactory.doc(s)))
    print 'The default solver is glpk.'
    print ""
    wrapper = textwrap.TextWrapper(replace_whitespace=False)
    print wrapper.fill('Subsolver options can be specified by with the solver name followed by colon and then the subsolver.  For example, the following specifies that the asl solver will be used:')
    print '   --asl:PICO'
    print wrapper.fill('This indicates that the asl solver will launch the PICO executable to perform optimization. Currently, no other solver supports this syntax.')


#
# Setup Pyomo execution environment
#
def setup_environment(options):
    #
    global start_time
    start_time = time.time()
    if not options.quiet:
        sys.stdout.write('[%8.2f] Setting up Pyomo environment\n' % 0.0)
        sys.stdout.flush()
    #
    # Setup memory tracker
    #
    if (pympler_available is True) and (options.profile_memory >= 1):
        global memory_tracker
        memory_tracker = tracker.SummaryTracker()
    #
    # Disable garbage collection
    #
    if options.disable_gc:
        gc.disable()
    #
    # Setup management for temporary files
    #
    if not options.tempdir is None:
        if not os.path.exists(options.tempdir):
            msg =  'Directory for temporary files does not exist: %s'
            raise ValueError, msg % options.tempdir
        TempfileManager.tempdir = options.tempdir

    #
    # Configure exception management
    #
    def pyomo_excepthook(etype,value,tb):
        """
        This exception hook gets called when debugging is on. Otherwise,
        run_command in this module is called.
        """
        global filter_excepthook
        if len(options.model_file) > 0:
            name = "model " + options.model_file
        else:
            name = "model"


        if filter_excepthook:
            action = "loading"
        else:
            action = "running"

        msg = "Unexpected exception while %s %s\n" % (action, name)

        #
        # This handles the case where the error is propagated by a KeyError.
        # KeyError likes to pass raw strings that don't handle newlines
        # (they translate "\n" to "\\n"), as well as tacking on single
        # quotes at either end of the error message. This undoes all that.
        #
        if etype == KeyError:
            valueStr = str(value).replace("\\n","\n")[1:-1]
        else:
            valueStr = str(value)

        logger.error(msg+valueStr)
        
        tb_list = traceback.extract_tb(tb,None)
        i = 0
        if not logger.isEnabledFor(logging.DEBUG) and filter_excepthook:
            while i < len(tb_list):
                #print "Y",model,tb_list[i][0]
                if options.model_file in tb_list[i][0]:
                    break
                i += 1
            if i == len(tb_list):
                i = 0
        print "\nTraceback (most recent call last):"
        for item in tb_list[i:]:
            print "  File \""+item[0]+"\", line "+str(item[1])+", in "+item[2]
            if item[3] is not None:
                print "    "+item[3]
        sys.exit(1)
    sys.excepthook = pyomo_excepthook
    # 
    return True


#
# Execute preprocessing files
#
def apply_preprocessing(options, parser):
    #
    if not options.quiet:
        sys.stdout.write('[%8.2f] Applying Pyomo preprocessing actions\n' % (time.time()-start_time))
        sys.stdout.flush()
    #
    global filter_excepthook
    #
    #
    # Setup solver and model
    #
    #
    if len(options.model_file) == 0:
        parser.print_help()
        return False
    #
    if not options.preprocess is None:
        for file in options.preprocess:
            preprocess = pyutilib.misc.import_file(file)
    #
    for ep in ExtensionPoint(IPyomoScriptPreprocess):
        ep.apply( options=options )
    #
    filter_excepthook=True
    options.usermodel = pyutilib.misc.import_file(options.model_file)
    filter_excepthook=False

    usermodel_dir = dir(options.usermodel)
    options._usermodel_plugins = []
    for key in modelapi:
        if key in usermodel_dir:
            class TMP(Plugin):
                implements(modelapi[key])
                def __init__(self):
                    Plugin.__init__(self)
                    self.fn = getattr(options.usermodel, key)
                def apply(self,**kwds):
                    return self.fn(**kwds)
            options._usermodel_plugins.append( TMP() )

    if 'pyomo_preprocess' in usermodel_dir:
        if options.model_name in usermodel_dir:
            msg = "Preprocessing function 'pyomo_preprocess' defined in file" \
                  " '%s', but model is already constructed!"
            raise SystemExit, msg % options.model_file
        getattr(options.usermodel, 'pyomo_preprocess')( options=options )

    return True

#
# Create instance of Pyomo model
#
def create_model(options):
    #
    if not options.quiet:
        sys.stdout.write('[%8.2f] Creating model\n' % (time.time()-start_time))
        sys.stdout.flush()
    #
    if (pympler_available is True) and (options.profile_memory >= 1):
        global memory_data
        objects_before_instance_creation = muppy.get_objects()
        memory_data.summary_before_instance_creation = summary.summarize(objects_before_instance_creation)
        print "Initial set of objects:"
        summary.print_(memory_data.summary_before_instance_creation,limit=50)
    #
    # Verify that files exist
    #
    for file in [options.model_file]+options.data_files:
        if not os.path.exists(file):
            raise IOError, "File "+file+" does not exist!"
    #
    # Create Model
    #
    ep = ExtensionPoint(IPyomoScriptCreateModel)
    model_name = 'model'
    if options.model_name is not None: model_name = options.model_name

    if model_name in dir(options.usermodel):
        if len(ep) > 0:
            msg = "Model construction function 'create_model' defined in "    \
                  "file '%s', but model is already constructed!"
            raise SystemExit, msg % options.model_file
        model = getattr(options.usermodel, model_name)

        if model is None:
            msg = "'%s' object is 'None' in module %s"
            raise SystemExit, msg % (model_name, options.model_file)
            sys.exit(0)

    else:
       if len(ep) == 0:
           msg = "Neither '%s' nor 'pyomo_create_model' are available in "    \
                 'module %s'
           raise SystemExit, msg % ( model_name, options.model_file )
       elif len(ep) > 1:
           msg = 'Multiple model construction plugins have been registered!'
           raise SystemExit, msg
       else:
            model_options = options.model_options
            if model_options is None:
                model_options = []
            model = ep.service().apply( options=pyutilib.misc.Container(*model_options) )
    #
    for ep in ExtensionPoint(IPyomoScriptPrintModel):
        ep.apply( options=options, model=model )

    #
    # Disable canonical repn for ASL solvers, and if the user has specified as such (in which case, we assume they know what they are doing!).
    #
    # Likely we need to change the framework so that canonical repn
    # is not assumed to be required by all solvers?
    #
    if not options.solver is None and options.solver.startswith('asl'):
        model.skip_canonical_repn = True
    elif options.skip_canonical_repn is True:
        model.skip_canonical_repn = True

    #
    # Create Problem Instance
    #
    ep = ExtensionPoint(IPyomoScriptCreateModelData)
    if len(ep) > 1:
        msg = 'Multiple model data construction plugins have been registered!'
        raise SystemExit, msg

    if len(ep) == 1:
        modeldata = ep.service().apply( options=options, model=model )
    else:
        modeldata = ModelData()

    if len(options.data_files) > 1:
        #
        # Load a list of *.dat files
        #
        for file in options.data_files:
            suffix = (file).split(".")[-1]
            if suffix != "dat":
                msg = 'When specifiying multiple data files, they must all '  \
                      'be *.dat files.  File specified: %s'
                raise SystemExit, msg % str( file )

            modeldata.add(file)

        modeldata.read(model)

        if not options.profile_memory is None:
           instance = model.create(modeldata, profile_memory=options.profile_memory)
        else:
           instance = model.create(modeldata)

    elif len(options.data_files) == 1:
       #
       # Load a *.dat file or process a *.py data file
       #
       suffix = (options.data_files[0]).split(".")[-1]
       if suffix == "dat":
          if not options.profile_memory is None:
             instance = model.create(options.data_files[0], profile_memory=options.profile_memory)
          else:
             instance = model.create(options.data_files[0])
       elif suffix == "py":
          userdata = pyutilib.misc.import_file(options.data_files[0])
          if "modeldata" in dir(userdata):
                if len(ep) == 1:
                    msg = "Cannot apply 'pyomo_create_modeldata' and use the" \
                          " 'modeldata' object that is provided in the model"
                    raise SystemExit, msg

                if userdata.modeldata is None:
                    msg = "'modeldata' object is 'None' in module %s"
                    raise SystemExit, msg % str( options.data_files[0] )

                modeldata=userdata.modeldata

          else:
                if len(ep) == 0:
                    msg = "Neither 'modeldata' nor 'pyomo_create_model_data"  \
                          'is defined in module %s'
                    raise SystemExit, msg % str( options.data_files[0] )

          modeldata.read(model)
          if not options.profile_memory is None:
             instance = model.create(modeldata, profile_memory=options.profile_memory)
          else:
             instance = model.create(modeldata)
       else:
          raise ValueError, "Unknown data file type: "+options.data_files[0]
    else:
       if not options.profile_memory is None:
          instance = model.create(modeldata, profile_memory=options.profile_memory)
       else:
          instance = model.create(modeldata)

    if options.linearize_expressions is True:
       linearize_model_expressions(instance)

    #
    ep = ExtensionPoint(IPyomoScriptModifyInstance)
    for ep in ExtensionPoint(IPyomoScriptModifyInstance):
        ep.apply( options=options, model=model, instance=instance )
    #
    if logger.isEnabledFor(logging.DEBUG):
       print "MODEL INSTANCE"
       instance.pprint()
       print ""

    for ep in ExtensionPoint(IPyomoScriptPrintInstance):
        ep.apply( options=options, instance=instance )

    fname=None
    symbol_map=None
    #if options.save_model is None and options.debug:
        #options.save_model = 'unknown.lp'
    if not options.save_model is None:
        if options.save_model == True:
            if options.format in (ProblemFormat.cpxlp, ProblemFormat.lpxlp):
                fname = (options.data_files[0])[:-3]+'lp'
            else:
                fname = (options.data_files[0])[:-3]+str(options.format)
            format=options.format
        else:
            fname = options.save_model
            format=None
        (fname, symbol_map) = instance.write(filename=fname, format=format)
        if not options.quiet:
            if not os.path.exists(fname):
                print "ERROR: file "+fname+" has not been created!"
            else:
                print "Model written to file '"+str(fname)+"'"
    for ep in ExtensionPoint(IPyomoScriptSaveInstance):
        ep.apply( options=options, instance=instance )

    if (pympler_available is True) and (options.profile_memory >= 1):
        objects_after_instance_creation = muppy.get_objects()
        memory_data.summary_after_instance_creation = summary.summarize(objects_after_instance_creation)
        print "\nObjects created during Pyomo instance creation:"
        memory_tracker.print_diff(
                    summary1=memory_data.summary_before_instance_creation, 
                    summary2=memory_data.summary_after_instance_creation)

    return pyutilib.misc.Container( 
                    model=model, instance=instance, 
                    symbol_map=symbol_map, filename=fname )

#
# Perform optimization with concrete instance
#
def apply_optimizer(options, instance):
    #
    if not options.quiet:
        sys.stdout.write('[%8.2f] Applying solver\n' % (time.time()-start_time))
        sys.stdout.flush()
    #
    #
    # Create Solver and Perform Optimization
    #
    solver = options.solver
    if solver is None:
       raise ValueError, "Problem constructing solver:  no solver specified"

    subsolver=None
    if not solver is None and ':' in solver:
        solver, subsolver = solver.split(':')
    opt = SolverFactory( solver )
    if opt is None:
        solver_list = SolverFactory.services()
        solver_list = sorted( filter(lambda x: '_' != x[0], solver_list) )
        raise ValueError, "Problem constructing solver `"+str(solver)+"' (choose from: %s)" % ", ".join(solver_list)

    # let the model know of our chosen solver's capabilities
    instance.has_capability = opt.has_capability

    opt.keepFiles=options.keepfiles or options.log
    if options.timelimit == 0:
       options.timelimit=None

    if not options.solver_suffixes is None:
        opt.suffixes = options.solver_suffixes

    if not subsolver is None:
        subsolver=' solver='+subsolver
    else:
        subsolver=''

    if not options.solver_options is None:
        opt.set_options(" ".join(options.solver_options)+subsolver)

    if options.smanager_type is None:
        solver_mngr = SolverManagerFactory( 'serial' )
    else:
        solver_mngr = SolverManagerFactory( options.smanager_type )

    if solver_mngr is None:
        msg = "Problem constructing solver manager '%s'"
        raise ValueError, msg % str( options.smanager_type )

    results = solver_mngr.solve( instance, opt=opt, tee=options.tee, timelimit=options.timelimit )

    if results == None:
       raise ValueError, "opt.solve returned None"

    if (pympler_available is True) and (options.profile_memory >= 1):
        global memory_data
        objects_after_optimization = muppy.get_objects()
        memory_data.summary_after_optimization = summary.summarize(objects_after_optimization)
        print "\nObjects created during optimization:"
        memory_tracker.print_diff(
                    summary1=memory_data.summary_after_instance_creation,
                    summary2=memory_data.summary_after_optimization)

    return results, opt

#
# Process results
#
def process_results(options, instance, results, opt):
    #
    if not options.quiet:
        sys.stdout.write('[%8.2f] Processing results\n' % (time.time()-start_time))
        sys.stdout.flush()
    #
    if options.log:
       print ""
       print "=========================================================="
       print "Solver Logfile:",opt.log_file
       print "=========================================================="
       print ""
       INPUT = open(opt.log_file, "r")
       for line in INPUT:
         print line,
       INPUT.close()
    #
    try:
        instance.update_results(results)
    except Exception, e:
        print "Problem loading solver results"
        raise
    #
    if not options.show_results:
        if options.save_results:
            results_file = options.save_results
        else:
            results_file = 'results.yml'
        results.write(filename=results_file)
        print "    Number of solutions:", len(results.solution)
        if len(results.solution) > 0:
            print "    Solution Information"
            print "      Gap:",results.solution[0].gap
            print "      Status:",results.solution[0].status
            if len(results.solution[0].objective) == 1:
                key = results.solution[0].objective.keys()[0]
                print "      Function Value:",results.solution[0].objective[key].value
        print "    Solver results file:",results_file
    #
    ep = ExtensionPoint(IPyomoScriptPrintResults)
    if len(ep) == 0:
        try:
            instance.load(results)
        except Exception, e:
            print "Problem loading solver results"
            raise
    if options.show_results:
        print ""
        results.write(num=1)
        print ""
    #
    if options.summary:
       print ""
       print "=========================================================="
       print "Solution Summary"
       print "=========================================================="
       if len(results.solution(0).variable) > 0:
          print ""
          display(instance)
          print ""
       else:
          print "No solutions reported by solver."
    #
    for ep in ExtensionPoint(IPyomoScriptPrintResults):
        ep.apply( options=options, instance=instance, results=results )
    #
    for ep in ExtensionPoint(IPyomoScriptSaveResults):
        ep.apply( options=options, instance=instance, results=results )
    #
    if (pympler_available is True) and (options.profile_memory >= 1):
        global memory_data
        objects_after_results_processing = muppy.get_objects()
        memory_data.summary_after_results_processing = summary.summarize(objects_after_results_processing)
        #diff_summary = summary.get_diff( memory_data.summary_after_optimization, memory_data.summary_after_results_processing)
        print "\nObjects created during results processing:"
        memory_tracker.print_diff(
                    summary1=memory_data.summary_after_optimization,
                    summary2=memory_data.summary_after_results_processing)

    return True


def apply_postprocessing(options, instance, results):
    #
    if not options.quiet:
        sys.stdout.write('[%8.2f] Applying Pyomo postprocessing actions\n' % (time.time()-start_time))
        sys.stdout.flush()
    #
    for file in options.postprocess:
        postprocess = pyutilib.misc.import_file(file)
        if "postprocess" in dir(postprocess):
            postprocess.postprocess(instance,results)
    for ep in ExtensionPoint(IPyomoScriptPostprocess):
        ep.apply( options=options, instance=instance, results=results )
    #
    # Deactivate and delete plugins
    #
    for plugin in options._usermodel_plugins:
        plugin.deactivate()
    options._usermodel_plugins = []
    #
    if (pympler_available is True) and (options.profile_memory >= 1):
        final_objects = muppy.get_objects()
        summary_final = summary.summarize(final_objects)
        final_summary = summary.summarize(final_objects)
        print "\nFinal set of objects:"
        summary.print_(final_summary, limit=50)
    #
    return True


def finalize(options, model=None, instance=None, results=None):
    #
    if not options.quiet:
        sys.stdout.write('[%8.2f] Pyomo Finished\n' % (time.time()-start_time))
        sys.stdout.flush()
    #
    model=model
    instance=instance
    results=results
    #
    if options.interactive:
        if IPython_available:
            ipshell()
        else:
            import code
            shell = code.InteractiveConsole(locals())
            print '\n# Dropping into Python interpreter'
            shell.interact()
            print '\n# Leaving Interpreter, back to Pyomo\n'


#
# Execute a function that processes command-line arguments and then
# calls a command-line driver.  This
# is segregated from the driver to enable profiling.
#
def run_command(command, parser, args=None, name='unknown'):
    #
    #
    # Parse command-line options
    #
    #
    try:
       _options = parser.parse_args(args=args)
       # Replace the parser options object with a pyutilib.misc.Options object
       options = pyutilib.misc.Options()
       for key in dir(_options):
            if key[0] != '_':
                val = getattr(_options, key)
                if not isinstance(val, types.MethodType):
                    options[key] = val
    except SystemExit:
       # the parser throws a system exit if "-h" is specified - catch
       # it to exit gracefully.
       return
    #
    # Configure the logger
    #
    logging.getLogger('coopr.pyomo').setLevel(logging.ERROR)
    logging.getLogger('coopr').setLevel(logging.ERROR)
    logging.getLogger('pyutilib').setLevel(logging.ERROR)
    #
    if options.warning:
        logging.getLogger('coopr.pyomo').setLevel(logging.WARNING)
        logging.getLogger('coopr').setLevel(logging.WARNING)
        logging.getLogger('pyutilib').setLevel(logging.WARNING)
    if options.info:
        logging.getLogger('coopr.pyomo').setLevel(logging.INFO)
        logging.getLogger('coopr').setLevel(logging.INFO)
        logging.getLogger('pyutilib').setLevel(logging.INFO)
    if options.verbose > 0:
        if options.verbose >= 1:
            logger.setLevel(logging.DEBUG)
        if options.verbose >= 2:
            logging.getLogger('coopr').setLevel(logging.DEBUG)
        if options.verbose >= 3:
            logging.getLogger('pyutilib').setLevel(logging.DEBUG)
    if options.debug:
        logging.getLogger('coopr.pyomo').setLevel(logging.DEBUG)
        logging.getLogger('coopr').setLevel(logging.DEBUG)
        logging.getLogger('pyutilib').setLevel(logging.DEBUG)
    if options.logfile:
        logging.getLogger('coopr.pyomo').handlers = []
        logging.getLogger('coopr').handlers = []
        logging.getLogger('pyutilib').handlers = []
        logging.getLogger('coopr.pyomo').addHandler( logging.FileHandler(options.logfile, 'w'))
        logging.getLogger('coopr').addHandler( logging.FileHandler(options.logfile, 'w'))
        logging.getLogger('pyutilib').addHandler( logging.FileHandler(options.logfile, 'w'))
    #
    # Setup I/O redirect to a file
    #
    logfile = getattr(options, 'output', None)
    if not logfile is None:
        pyutilib.misc.setup_redirect(logfile)
    #
    # Call the main Pyomo runner with profiling
    #
    if options.profile > 0:
        if not pstats_available:
            if not logfile is None:
                pyutilib.misc.reset_redirect()
            msg = "Cannot use the 'profile' option.  The Python 'pstats' "    \
                  'package cannot be imported!'
            raise ValueError, msg
        tfile = TempfileManager.create_tempfile(suffix=".profile")
        tmp = profile.runctx(
          command.__name__ + '(options=options,parser=parser)', command.__globals__, locals(), tfile
        )
        p = pstats.Stats(tfile).strip_dirs()
        p.sort_stats('time', 'cum')
        p = p.print_stats(options.profile)
        p.print_callers(options.profile)
        p.print_callees(options.profile)
        p = p.sort_stats('cum','calls')
        p.print_stats(options.profile)
        p.print_callers(options.profile)
        p.print_callees(options.profile)
        p = p.sort_stats('calls')
        p.print_stats(options.profile)
        p.print_callers(options.profile)
        p.print_callees(options.profile)
        TempfileManager.clear_tempfiles()
        ans = [tmp, None]
    else:
        #
        # Call the main Pyomo runner without profiling
        #
        try:
            ans = command(options=options, parser=parser)
        except SystemExit, err:
            if __debug__:
                if options.debug or options.catch:
                    sys.exit(0)
            print 'Exiting %s: %s' % (name, str(err))
            ans = None
        except Exception, err:
            # If debugging is enabled, pass the exception up the chain
            # (to pyomo_excepthook)
            if __debug__:
                if options.debug or options.catch:
                    if not logfile is None:
                        pyutilib.misc.reset_redirect()
                    raise

            if len(options.model_file) > 0:
                name = "model " + options.model_file
            else:
                name = "model"
                
            global filter_excepthook
            if filter_excepthook:
                action = "loading"
            else:
                action = "running"

            msg = "Unexpected exception while %s %s\n" % (action, name)
            #
            # This handles the case where the error is propagated by a KeyError.
            # KeyError likes to pass raw strings that don't handle newlines
            # (they translate "\n" to "\\n"), as well as tacking on single
            # quotes at either end of the error message. This undoes all that.
            #
            errStr = str(err)
            if type(err) == KeyError:
                errStr = str(err).replace(r"\n","\n")[1:-1]

            logging.getLogger('coopr.pyomo').error(msg+errStr)
            ans = None

    if not logfile is None:
        pyutilib.misc.reset_redirect()

    if options.disable_gc:
        gc.enable()
    return ans



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['RangeSet']

import itertools
import sets
import expr
from numvalue import value
from pyutilib.component.core import *

def count(start=0, step=1):
    n = start
    while True:
        yield n
        n += step

class RangeSetValidator(object):

    def __init__(self, start, end, step):
        self.start=start
        self.end=end
        self.step=step

    def __call__(self, val, model):
       if not type(val) in [int, float, long]:
            return False
       if val + 1e-7 < self.start:
          return False
       if val > self.end+1e-7:
          return False
       if type(self.start) is int and type(self.end) is int and type(self.step) is int and (val-self.start)%self.step != 0:
          return False
       return True


class RangeSet(sets._SetContainer):
    """A set that represents a list of integer values"""

    alias("RangeSet", "A sequence of numeric values.  RangeSet(start,end,step) is a sequence starting a value 'start', and increasing in values by 'step' until a value greater than of equal to 'end' is reached.")

    def __init__(self,*args,**kwds):
        """Construct a list of integers"""
        tmp=()
        sets._SetContainer.__init__(self,*tmp,**kwds)
        self._type=RangeSet
        self.ordered=True
        if len(args) == 1:
           self._start=1
           self._end=args[0]
           self._step=1
        elif len(args) == 2:
           self._start=args[0]
           self._end=args[1]
           self._step=1
        else:
           self._start=args[0]
           self._end=args[1]
           self._step=args[2]

    def construct(self, values=None):
        if self._constructed:
            return
        self._constructed=True
        if isinstance(self._start,expr.Expression):
           self._start_val = self._start()
        else:
           self._start_val = value(self._start)

        if isinstance(self._end,expr.Expression):
           self._end_val = self._end()
        else:
           self._end_val = value(self._end)

        if isinstance(self._step,expr.Expression):
           self._step_val = self._step()
        else:
           self._step_val = value(self._step)

        if self.validate is None:
           self.validate=RangeSetValidator(self._start_val, self._end_val, self._step_val)

        #
        # Generate set, and track the bounds
        #
        lb = self._start_val
        ub = self._start_val

        for i in itertools.islice(count(), (self._end_val-self._start_val+self._step_val+1e-7)//self._step_val): 
          val = self._start_val + i*self._step_val
          if self._verify(val,False):
             ub=val
             self.add(val)

        #
        # Apply the filter, if present
        #
        if self.filter is not None:
            self._apply_filter()
            lb = min(self.value)
            ub = max(self.value)

        # Assign bounds
        self._bounds=(lb,ub)



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2010 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the FAST README.txt file.
#  _________________________________________________________________________

__all__ = ['linearize_model_expressions', 'is_linear_expression_constant']

# a prototype utility to translate from a "normal" model instance containing
# constraints and objectives with expresssions for their bodies, into a 
# more compact linearized verson of the problem. eliminates the need for
# canonical expressions and the associated overhead (memory and speed).

# NOTE: Currently only works for constraints - leaving the expressions alone,
#       mainly because PH doesn't modify those and I don't want to deal with
#       that aspect for the moment. In particular, this only works for 
#       immutable parameters due to the reliance on the canonical expression
#       generator to extract the linear and constant terms.

from coopr.pyomo import *
from coopr.pyomo.expr import generate_canonical_repn, is_linear, is_constant

def linearize_model_expressions(instance):

   # TBD: Should we really be doing this for all components, and not just active ones?
   for constraint_name, constraint in instance.active_components()[Constraint].iteritems():

      for index, constraint_data in constraint._data.iteritems():

         if constraint_data.repn != None:
            canonical_encoding = constraint_data.repn
         else:
            canonical_encoding = generate_canonical_repn(constraint_data.body)

         variable_map = canonical_encoding[-1]

         # we obviously can't linearize an expression if it has higher-order terms!
         if is_linear(canonical_encoding) or is_constant(canonical_encoding):

            constant_term = 0.0
            linear_terms = [] # a list of coefficient, _VarValue pairs.

            if canonical_encoding.has_key(0):
               constant_term = canonical_encoding[0][None]

            for variable_key, coefficient in canonical_encoding[1].iteritems():
               var_value = variable_map[variable_key.keys()[0]]
               linear_terms.append((coefficient, var_value))

            # eliminate the expression tree - we don't need it any longer.
            # ditto the canonical representation.
            constraint_data.body = None
            constraint_data.lin_body = [constant_term, linear_terms]
            if constraint_data.repn is not None:
               constraint_data.repn = None

def is_linear_expression_constant(lin_body):
    """Return True if the linear expression is a constant expression, due either to a lack of variables or fixed variables"""
    return (len(lin_body[1]) == 0)



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2010 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the FAST README.txt file.
#  _________________________________________________________________________

# this module provides some basic functionality for generating labels from pyomo
# names, which often contain characters such as "[" and "]" (e.g., in my_var[1]).
# these characters generally cause issues with optimization input file formats,
# e.g., CPLEX LP files. the purpose of this module is to provide a simple remap
# function, that will satisfy broadly problematic symbols. if solver-specific 
# remaps are required, they should be handled either in a preprocessor or a solver
# plugin.

__all__ = ['label_from_name']

ignore = set(['"', "'", '`'])
remap = {}
# Remap as underscores
remap['-']='_'
remap['_']='_'
remap[' ']='_'
remap['(']='['
remap['{']='['
remap[')']=']'
remap['}']=']'
# Remap as Q+ASCII-Dec
remap['!']='_Q33_'
remap['#']='_Q35_'
remap['$']='_Q36_'
remap['%']='_Q37_'
remap['&']='_Q38_'
remap['*']='_Q42_'
remap['+']='_Q43_'
#remap[',']='_Q44_'
remap['.']='_Q46_'
remap['/']='_Q47_'
remap[';']='_Q59_'
remap['<']='_Q60_'
remap['=']='_Q61_'
remap['>']='_Q62_'
remap['?']='_Q63_'
remap['@']='_Q63_'
remap['\\']='_Q92_'
remap['^']='_Q94_'
remap['|']='_Q124_'
remap['~']='_Q126_'

def label_from_name(name):

   global ignore
   global remap

   if name is None:
      raise RuntimeError, "Illegal name=None supplied to label_from_name function"

   label=""
   for c in name:
      if c in remap:
         label += remap[c]
      elif not c in ignore:
         label += c
   return label



#
# Test the Pyomo command-line interface
#

import unittest
import os
import sys
from os.path import abspath, dirname
currdir = dirname(abspath(__file__))+os.sep
scriptdir = dirname(dirname(dirname(dirname(dirname(abspath(__file__))))))+os.sep

import coopr.pyomo
import pyutilib.services
import pyutilib.subprocess
import pyutilib.th as unittest
import coopr.pyomo.scripting.pyomo as main
import StringIO 
from pyutilib.misc import setup_redirect, reset_redirect
import re

if os.path.exists(sys.exec_prefix+os.sep+'bin'+os.sep+'coverage'):
    executable=sys.exec_prefix+os.sep+'bin'+os.sep+'coverage -x '
else:
    executable=sys.executable

def filter_fn(line):
    return line.startswith('Disjunct')


class Test(unittest.TestCase):

    def pyomo(self, cmd, **kwds):
        args=re.split('[ ]+',cmd)
        if 'root' in kwds:
            OUTPUT=kwds['root']+'.out'
            results=kwds['root']+'.yml'
            self.ofile = OUTPUT
        else:
            OUTPUT=StringIO.StringIO()
            results='results.yml'
        setup_redirect(OUTPUT)
        os.chdir(currdir)
        output = main.run(['--save-results=%s' % results] + list(args))
        reset_redirect()
        if not 'root' in kwds:
            return OUTPUT.getvalue()
        return output

    def setUp(self):
        self.ofile = None

    def tearDown(self):
        if self.ofile and os.path.exists(self.ofile):
            os.remove(self.ofile)
        if os.path.exists(currdir+'results.yml'):
            os.remove(currdir+'results.yml')

    def run_pyomo(self, cmd, root=None):
        return pyutilib.subprocess.run('pyomo --save-results=%s.yml ' % (root) +cmd, outfile=root+'.out')
        
    def test1(self):
        """Simple execution of 'pyomo'"""
        self.pyomo('pmedian.py pmedian.dat', root=currdir+'test1')
        self.assertFileEqualsBaseline(currdir+"test1.yml", currdir+"test1.txt")

    def test1a(self):
        """Simple execution of 'pyomo'"""
        self.run_pyomo('pmedian.py pmedian.dat', root=currdir+'test1a')
        self.assertFileEqualsBaseline(currdir+"test1a.yml", currdir+"test1.txt")

    def test2(self):
        """Run pyomo with bad --model-name option value"""
        self.pyomo('--model-name=dummy pmedian.py pmedian.dat', root=currdir+'test2')
        def filter(line):
            return line.startswith('[')
        self.assertFileEqualsBaseline(currdir+"test2.out", currdir+"test2.txt", filter=filter)

    def test3(self):
        """Run pyomo with model that does not define model object"""
        self.pyomo('pmedian1.py pmedian.dat', root=currdir+'test3')
        def filter(line):
            return line.startswith('[')
        self.assertFileEqualsBaseline(currdir+"test3.out", currdir+"test3.txt", filter=filter)

    def test4(self):
        """Run pyomo with good --model-name option value"""
        self.pyomo('--model-name=MODEL pmedian1.py pmedian.dat', root=currdir+'test4')
        self.assertFileEqualsBaseline(currdir+"test4.yml", currdir+"test1.txt")

    def test5(self):
        """Run pyomo with create_model function"""
        self.pyomo('pmedian2.py pmedian.dat', root=currdir+'test5')
        def filter(line):
            return line.startswith("Writingmodel") or line.startswith("Solverresultsfile") or \
                   line.startswith('[')
        self.assertFileEqualsBaseline(currdir+"test5.out", currdir+"test5.txt", filter=filter)
        os.remove(currdir+'test5.yml')

    def test6(self):
        """Run pyomo with help-components option"""
        self.pyomo('--help-components', root=currdir+'test6')
        self.assertFileEqualsBaseline(currdir+"test6.out", currdir+"test6.txt", filter=filter_fn)

    def Xtest7(self):
        """Run pyomo with help option"""
        self.pyomo('--help', root=currdir+'test7')
        self.assertFileEqualsBaseline(currdir+"test7.yml", currdir+"test7.txt")

    def test8(self):
        """Run pyomo with --instance-only option"""
        output = self.pyomo('--instance-only pmedian.py pmedian.dat', root=currdir+'test8')
        self.assertEqual(type(output.instance), coopr.pyomo.AbstractModel)
        # Check that the results file was NOT created
        self.assertRaises(OSError, lambda: os.remove(currdir+'test8.yml'))

    def test9(self):
        """Run pyomo with --disable-gc option"""
        output = self.pyomo('--disable-gc pmedian.py pmedian.dat', root=currdir+'test9')
        self.assertEqual(type(output.instance), coopr.pyomo.AbstractModel)
        os.remove(currdir+'test9.yml')

    def test10(self):
        """Run pyomo with --verbose option"""
        def filter(line):
            return line.startswith("Writingmodel") or line.startswith("Solverresultsfile") or \
                   line.startswith('DEBUG:') or line.startswith('INFO:') or \
                   line.startswith('[' or line.endswith('cpxlp'))
        self.pyomo('-v pmedian.py pmedian.dat', root=currdir+'test10')
        self.assertFileEqualsBaseline(currdir+"test10.out", currdir+"test10.txt", filter)
        os.remove(currdir+'test10.yml')

    def Xtest11(self):
        """Run pyomo with --debug=generate option"""
        self.pyomo('--debug=generate pmedian.py pmedian.dat', root=currdir+'test11')
        self.assertFileEqualsBaseline(currdir+"test11.yml", currdir+"test11.txt")

    def test12(self):
        """Run pyomo with --output option"""
        def filter(line):
            #print "HERE",line
            return line.startswith("Writing")
        self.pyomo('--output=%s pmedian.py pmedian.dat' % (currdir+'test12.log'), root=currdir+'test12')
        self.assertFileEqualsBaseline(currdir+"test12.yml", currdir+"test12.txt", filter)
        os.remove(currdir+'test12.log')

    def test13(self):
        """Simple execution of 'pyomo' with implicit rules"""
        self.pyomo('pmedian3.py pmedian.dat', root=currdir+'test13')
        self.assertFileEqualsBaseline(currdir+"test13.yml", currdir+"test1.txt")


Test = unittest.skipIf(pyutilib.services.registered_executable("glpsol") is None, "The 'glpsol' executable is not available")(Test)

if __name__ == "__main__":
   unittest.main()



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

"""
Block Properties:

1. Blocks are indexed components that contain other components (including blocks)

2. Blocks have a global attribute that defines whether construction is deferred.  This 
applies to all components that they contain except blocks.  Blocks contained by other blocks
use their local attribute to determine whether construction is deferred.

NOTE: Blocks do not currently maintain statistics about the sets, parameters, constraints, etc that
they contain, including these components from subblocks.
"""

__all__ = ['Block']

import sys

from plugin import *

from pyutilib.component.core import alias, ExtensionPoint
from pyutilib.misc import Container

from component import Component
from sets import Set, _ProductSet, _SetContainer, _BaseSet
from rangeset import RangeSet
from var import Var
from constraint import Objective, Constraint
from param import Param

from indexed_component import IndexedComponent
try:
    from components import OrderedDict
except ImportError:
    from ordereddict import OrderedDict


class BlockComponents(object):

    def __init__(self):
        #
        # Define component dictionary: component type -> instance
        #
        self._component={}
        for item in ModelComponentFactory.services():
            self._component[  ModelComponentFactory.get_class(item) ] = {}
        #
        # A list of the declarations, in the order that they are
        # specified.
        #
        self._declarations=OrderedDict()

    def __iter__(self):
        return self._declarations.iterkeys()

    def __getitem__(self, name):
        if isinstance(name,basestring):
            return self._declarations.get(name, None)
        return self._component.get(name, None)

    def _clear_attribute(self, name):
        val = self[name]
        if val is None:
            return

        if isinstance(val, Block):
            val.__dict__['_parent_block'] = None
            val.model = val
        else:
            val.model = None
            
        del self._component[ val.type() ][name]
        del self._declarations[name]
        self.__dict__[name]=None

    def _add_component(self, name, val):
        self._component[val.type()][name]=val
        self._declarations[name] = val
        self.__dict__[name]=val

    def __setattr__(self,name,val):
            #
            # Try to set the value. This may fail if the attribute
            # does not already exist in this model, if the set_value
            # function is not defined, or if a bad value is
            # provided. In the latter case, a ValueError will be
            # thrown, which # we raise. Otherwise, this is an object
            # that we need to set directly.
            #
            try:
                self.__dict__[name].set_value(val)
            except ValueError, e:
                raise
            except Exception, e:
                self.__dict__[name]=val

    def iterkeys(self):
        return self._declarations.iterkeys()

    def itervalues(self):
        return self._declarations.itervalues()

    def iteritems(self):
        return self._declarations.iteritems()

    def components(self, ctype=None):
        """
        Return information about the block components.  If ctype is None, return the dictionary
        that maps {component type -> {name -> instance}}.  Otherwise, return the dictionary
        that maps {name -> instance} for the specified component type.
        """
        if ctype is None:
            return self._component
        if ctype in self._component:
             return self._component[ctype]
        raise KeyError, "Unknown component type: %s" % str(ctype)

    def active_components(self, _ctype=None):
        """
        Returns the active components in this block.  If _ctype is None, return the
        dictionary that maps {component type -> {name -> instance}}.  Otherwise, return
        the dictionary that maps {name -> instance} for the specified component type.
        """
        tmp = {}
        if _ctype is None:
            for ctype in self._component:
                tmp[ctype]={}
        elif _ctype in self._component:
            tmp[_ctype]={}
        else:
            raise KeyError, "Unknown component type: %s" % str(_ctype)
        for ctype in tmp:
            for name in self._component[ctype]:
                comp = self._component[ctype][name]
                if comp.active:
                    tmp[ctype][name] = comp
        if not _ctype is None:
            return tmp[_ctype]
        return tmp

    def is_constructed(self):
       """
       A boolean indicating whether or not all *active* components of the
       input model have been properly constructed.
       """
       component_map = arg.active_components()
       for type, entries in component_map.iteritems():
          for component_name, component in entries.iteritems():
             if component.is_constructed() is False:
                return False
       return True

    def pprint(self, filename=None, ostream=None):
        """
        Print a summary of the model info
        """
        if ostream is None:
           ostream = sys.stdout
        if filename is not None:
           OUTPUT=open(filename,"w")
           self.pprint(ostream=OUTPUT)
           OUTPUT.close()
           return
        if ostream is None:
           ostream = sys.stdout
        #
        # We hard-code the order of the core Pyomo modeling
        # components, to ensure that the output follows the logical order
        # that expected by a user.
        #
        items = [Set, RangeSet, Param, Var, Objective, Constraint, Block]
        for item in ExtensionPoint(IModelComponent):
            if not item in items:
                items.append(item)

        # Currently, model components are not actually being registered
        # with the IModelComponent extension point (1 Nov 2010), so as a
        # workaround, we will loop through the components and add any
        # new components that actually have defined members.
        extra_items = []
        for item, members in self._component.iteritems():
            if item not in items and len(members):
                extra_items.append(item)
        # extra items get added alphabetically (so output is consistent)
        items.extend(sorted(extra_items))

        for item in items:
            if not item in self._component:
                continue
            keys = self._component[item].keys()
            keys.sort()
            #
            # NOTE: these conditional checks should not be hard-coded.
            #
            print >>ostream, len(keys), item.__name__+" Declarations"
            for key in keys:
                self._component[item][key].pprint(ostream=ostream)
            print >>ostream, ""
        #
        # Model Order
        #
        print >>ostream, len(self._declarations),"Declarations:",
        for name in self._declarations:
          print >>ostream, name,
        print >>ostream, ""


class Block(IndexedComponent):
    """
    A block in an optimization model.  By default, this defers construction of 
    components until data is loaded.
    """

    alias("Block", "Blocks are indexed components that contain one or more " + \
          "other model components.")

    def __init__(self, *args, **kwargs):
        """Constructor"""
        if 'ctype' in kwargs:
            tkwargs = {'ctype':kwargs['ctype']}
        else:
            tkwargs = {'ctype':Block}
        IndexedComponent.__init__(self, *args, **tkwargs)
        #
        self.name=kwargs.get('name', 'unknown')
        self._defer_construction=True
        # By default, a Block is a top-level (i.e. Model) block
        self.model = self
        self._parent_block = None
        self.components =  BlockComponents()

    def concrete_mode(self):
        """Configure block to immediately construct components"""
        self._defer_construction=False

    def symbolic_mode(self):
        """Configure block to defer construction of components"""
        self._defer_construction=True

    def components(self, ctype=None):
        """
        Return information about the block components.  If ctype is None, return the dictionary
        that maps {component type -> {name -> instance}}.  Otherwise, return the dictionary
        that maps {name -> instance} for the specified component type.
        """
        return self.components.components(ctype)

    def active_components(self, _ctype=None):
        """
        Returns the active components in this block.  If _ctype is None, return the
        dictionary that maps {component type -> {name -> instance}}.  Otherwise, return
        the dictionary that maps {name -> instance} for the specified component type.
        """
        return self.components.active_components(_ctype)
        
    def _add_temporary_set(self,val):
        if val._index_set is not None:
            ctr=0
            for tset in val._index_set:
                if tset.name == "_unknown_":
                    self._construct_temporary_set(
                      tset,
                      val.name+"_index_"+str(ctr)
                    )
                ctr += 1
            val._index = self._construct_temporary_set(
              val._index_set,
              val.name+"_index"
            )
        if isinstance(val._index,_SetContainer) and \
            val._index.name == "_unknown_":
            self._construct_temporary_set(val._index,val.name+"_index")
        if getattr(val,'domain',None) is not None and val.domain.name == "_unknown_":
            self._construct_temporary_set(val.domain,val.name+"_domain")

    def _construct_temporary_set(self, obj, name):
        if type(obj) is tuple:
           if len(obj) == 1:                #pragma:nocover
                  raise Exception, "Unexpected temporary set construction"
           else:
                  tobj=_ProductSet(*obj)
                  setattr(self,name,tobj)
                  tobj.virtual=True
                  return tobj
        if isinstance(obj,_BaseSet):
           setattr(self,name,obj)
           return obj

    def _clear_attribute(self, name, only_block=False):
        """
        Cleanup the pre-existing model attribute
        """
        #
        # We don't delete here because that would destruct the component, which might
        # be shared with another block.
        #
        self.__dict__[name]=None
        if only_block:
            return

        self.components._clear_attribute(name)

    def _add_component(self, name, val):
        self._clear_attribute(name)

        # all Pyomo components have names. but not all have labels (e.g., sets, params).
        # if it does have a label, sync the default label with the new name.
        val.name=name
        if hasattr(val,'label') is True:
           val.sync_label()

        self._add_temporary_set(val)
        self.components._add_component(name,val)
        self.__dict__[name]=val

        # Presumably self.model refers to the 'root' Block that will
        # eventually be solved, whereas '_parent_block' refers to the
        # immediate parent.

        # Update the (new) child block's _parent_block.  
        if isinstance(val, Block):
            if val._parent_block is not None:
                # Nothing really wrong here, but until we are more
                # careful about checking that this actually works,
                # complain loudly.
                raise Exception, "Reassigning a block attached to a model"
            # NB: use __dict__ directly to prevent infinite recursion
            val.__dict__['_parent_block'] = self

        # Update the new component's model pointer
        val.model = self.model
        
        frame = sys._getframe(2)
        locals_ = frame.f_locals
        if getattr(val,'rule',None) is None and val.name+'_rule' in locals_:
            val.rule = locals_[val.name+'_rule']
        if not self._defer_construction:
            val.construct(None)

    def __setattr__(self,name,val):
        """Set attributes"""
        #
        # Set Model Declaration
        #
        if name == "model":
            if val is not None and not isinstance(val, Block):
                raise ValueError, "Cannot set Block.model to a non-Block " \
                      "object (%s)" % ( val )
            self.__dict__["model"] = val
            # Update all child components.  NB: use __dict__.get because
            # Component.__init__() assigns "self.model = None" before
            # Block.__init__ defines the _declarations map.
            components = self.__dict__.get('components', None)
            if not components is None:
                for subcomp in components.itervalues():
                    subcomp.model = val
        elif name != "components" and isinstance(val, Component):
            #
            # If this is a component type, then simply set it
            #
            self._add_component(name,val)
        else:
            #
            # Try to set the value. This may fail if the attribute
            # does not already exist in this model, if the set_value
            # function is not defined, or if a bad value is
            # provided. In the latter case, a ValueError will be
            # thrown, which # we raise. Otherwise, this is an object
            # that we need to set directly.
            #
            try:
                self.__dict__[name].set_value(val)
            except ValueError, e:
                raise
            except Exception, e:
                self.__dict__[name]=val

    def pprint(self, filename=None, ostream=None):
        """
        Print a summary of the model info
        """
        self.components.pprint(filename=filename, ostream=ostream)

    def display(self, filename=None, ostream=None):
        """
        Print the Pyomo model in a verbose format.
        """
        if filename is not None:
           OUTPUT=open(filename,"w")
           self.display(ostream=OUTPUT)
           OUTPUT.close()
           return
        if ostream is None:
           ostream = sys.stdout
        if self.type() is Block:
            print >>ostream, "Block "+self.name
        else:
            print >>ostream, "Model "+self.name
        #
        print >>ostream, ""
        print >>ostream, "  Variables:"
        VAR = self.active_components(Var)
        if len(VAR) == 0:
            print >>ostream, "    None"
        else:
            for ndx in VAR:
                VAR[ndx].display(prefix="    ",ostream=ostream)
        #
        print >>ostream, ""
        print >>ostream, "  Objectives:"
        OBJ = self.active_components(Objective)
        if len(OBJ) == 0:
            print >>ostream, "    None"
        else:
            for ndx in OBJ:
                OBJ[ndx].display(prefix="    ",ostream=ostream)
        print >>ostream, ""
        #
        CON = self.active_components(Constraint)
        print >>ostream, "  Constraints:"
        if len(CON) == 0:
            print >>ostream, "    None"
        else:
            for ndx in CON:
                CON[ndx].display(prefix="    ",ostream=ostream)




#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

#
# AMPL Problem Writer Plugin
#

from coopr.opt import ProblemFormat, WriterFactory
from coopr.pyomo.base.numtypes import minimize, maximize
from coopr.pyomo.base import *
from coopr.pyomo.base import expr
from coopr.pyomo.base import intrinsic_functions
from coopr.pyomo.base.var import _VarValue, Var, _VarElement, _VarBase
from coopr.pyomo.base import numvalue
from coopr.opt.base import *
from coopr.pyomo.base.param import _ParamValue
from coopr.pyomo.base import var 
from coopr.pyomo.base import param 
from pyutilib.component.core import alias
from coopr.pyomo.expr.canonical_repn import is_nonlinear, is_linear

import logging
logger = logging.getLogger('coopr.pyomo')

class ProblemWriter_nl(AbstractProblemWriter):

    alias(str(ProblemFormat.nl))

    def __init__(self):
        AbstractProblemWriter.__init__(self,ProblemFormat.nl)

    def __call__(self, model, filename):
        if filename is None:
           filename = model.name + ".nl"
        OUTPUT=open(filename,"w")
        symbol_map = self._print_model_NL(model,OUTPUT)
        OUTPUT.close()
        return filename, symbol_map

    def _get_bound(self, exp):
        if isinstance(exp,expr._IdentityExpression):
           return self._get_bound(exp._args[0])
        elif isinstance(exp,NumericConstant):
           return exp.value
        elif isinstance(exp,_ParamValue):
           return exp.value
        else:
           raise ValueError, "ERROR: nonconstant bound: " + str(exp)
           return None
    
    def _print_nonlinear_terms_NL(self, OUTPUT, exp):
        if isinstance(exp, expr._IntrinsicFunctionExpression):
            if exp.name == 'log':
                print >>OUTPUT, "o43", " #log"
            elif exp.name == 'log10':
                print >>OUTPUT, "o42", " #log10"
            elif exp.name == 'sin':
                print >>OUTPUT, "o41", " #sin"
            elif exp.name == 'cos':
                print >>OUTPUT, "o46", " #cos"
            elif exp.name == 'tan':
                print >>OUTPUT, "o38", " #tan"
            elif exp.name == 'sinh':
                print >>OUTPUT, "o40", " #sinh"
            elif exp.name == 'cosh':
                print >>OUTPUT, "o45", " #cosh"
            elif exp.name == 'tanh':
                print >>OUTPUT, "o37", " #tanh"
            elif exp.name == 'asin':
                print >>OUTPUT, "o51", " #asin"
            elif exp.name == 'acos':
                print >>OUTPUT, "o53", " #acos"
            elif exp.name == 'atan':
                print >>OUTPUT, "o49", " #atan"
            elif exp.name == 'exp':
                print >>OUTPUT, "o44", " #exp"
            elif exp.name == 'sqrt':
                print >>OUTPUT, "o39", " #sqrt"
            elif exp.name == 'asinh':
                print >>OUTPUT, "o50", " #asinh"
            elif exp.name == 'acosh':
                print >>OUTPUT, "o52", " #acosh"
            elif exp.name == 'atanh':
                print >>OUTPUT, "o47", " #atanh"
            elif exp.name == 'pow':
                print >>OUTPUT, "o5", " #^"
            else:
                logger.error("Unsupported intrinsic function (%s)", exp.name)
                raise TypeError("ASL writer does not support '%s' expressions"
                                % exp.name)
                
            for child_exp in exp._args:
                self._print_nonlinear_terms_NL(OUTPUT, child_exp)
        
        elif isinstance(exp, expr._SumExpression):
            n = len(exp._args)
            if exp._const != 0.0:
                print >>OUTPUT, "o0", " #+"
                print >>OUTPUT, "n" + str(exp._const)
                
            for i in xrange(0,n-1):
                print >>OUTPUT, "o0", " #+"
                if exp._coef[i] != 1:
                    print >>OUTPUT, "o2", " #*"
                    print >>OUTPUT, "n" + str(exp._coef[i])
                self._print_nonlinear_terms_NL(OUTPUT, exp._args[i])

            if exp._coef[n-1] != 1:
                print >>OUTPUT, "o2", " #*"
                print >>OUTPUT, "n" + str(exp._coef[n-1])
            self._print_nonlinear_terms_NL(OUTPUT, exp._args[n-1])

        elif isinstance(exp, list):
            # this is an implied summation of expressions (we did not create a new sum expression for efficiency)
            # this should be a list of tuples where [0] is the coeff and [1] is the expr to write
            n = len(exp)
            for i in xrange(0,n):
                assert(isinstance(exp[i],tuple))
                coef = exp[i][0]
                child_exp = exp[i][1]
                if i != n-1:
                    # need the + op if it is not the last entry in the list
                    print >>OUTPUT, "o0", " #+"
                    
                if coef != 1.0: 
                    print >>OUTPUT, "o2", " #*"
                    print >>OUTPUT, "n" + str(coef)
                self._print_nonlinear_terms_NL(OUTPUT, child_exp)

        elif isinstance(exp, expr._ProductExpression):
            denom_exists = False
            if len(exp._denominator) == 0:
                pass
            else:
                print >>OUTPUT, "o3 # /"
                denom_exists = True
            
            if exp.coef != 1:
                print >>OUTPUT, "o2 # *"
                print >>OUTPUT, "n" + str(exp.coef)

            if len(exp._numerator) == 0:
                print >>OUTPUT, "n1"

            # print out the numerator
            child_counter = 0
            max_count = len(exp._numerator)-1
            for child_exp in exp._numerator:
#                if child_exp != exp._numerator[-1]:
                if child_counter < max_count:
                    print >>OUTPUT, "o2", " #*"
                self._print_nonlinear_terms_NL(OUTPUT, child_exp)
                child_counter += 1

            if denom_exists:
            # print out the denominator            
                child_counter = 0
                max_count = len(exp._denominator)-1
                for child_exp in exp._denominator:
                    if child_counter < max_count:
                        print >>OUTPUT, "o2", " #*"
                    self._print_nonlinear_terms_NL(OUTPUT, child_exp)
                    child_counter += 1

        elif isinstance(exp, expr._PowExpression):
            print >>OUTPUT, "o5", " #^"
            for child_exp in exp._args:
                self._print_nonlinear_terms_NL(OUTPUT, child_exp)

        elif (isinstance(exp,var._VarValue) or isinstance(exp,var.Var)) and not exp.fixed_value():
            print >>OUTPUT, "v" + str(exp.ampl_var_id) + " #" + str(exp.label)
            
        elif isinstance(exp,param._ParamValue):
            print >>OUTPUT, "n" + str(exp.value) + " #" + str(exp.name)

        elif (isinstance(exp,numvalue.NumericConstant) or exp.fixed_value):
            print >>OUTPUT, "n" + str(exp.value) + " # numeric constant"

        else:
            raise ValueError, "Unsupported expression type in _print_nonlinear_terms_NL"
            
    def _print_model_NL(self, model, OUTPUT, verbose=False):

        # maps NL variables to the "real" variable names in the problem.
        # it's really NL variable ordering, as there are no variable names
        # in the NL format. however, we by convention make them go from
        # x0 upward.
        symbol_map = {}

        #
        # Collect statistics
        #
        Obj = model.active_components(Objective)
        Con = model.active_components(Constraint)
        Vars = dict() # will be the entire list of all vars used in the problem
        #
        # Count number of objectives and build the ampl_repns
        #
        n_objs = 0
        n_nonlinear_objs = 0
        ObjVars = set()
        ObjNonlinearVars = set()
        for obj in Obj:
            if Obj[obj].trivial:
                continue
            for i in Obj[obj]:
                ampl_repn = generate_ampl_repn(Obj[obj][i].expr)
#                ampl_repn._nonlinear_expr.pprint()
                Obj[obj][i].ampl_repn = ampl_repn
                
                Vars.update(ampl_repn._linear_terms_var)
                Vars.update(ampl_repn._nonlinear_vars)
                
                ObjVars.update(ampl_repn._linear_terms_var.keys())
                ObjVars.update(ampl_repn._nonlinear_vars.keys())
                ObjNonlinearVars.update(ampl_repn._nonlinear_vars.keys())
                n_objs += 1
                if ampl_repn.is_nonlinear():
                    n_nonlinear_objs += 1
        if n_objs != 1:
            raise ValueError, "asl solver has detected multiple objective functions, but currently only handles a single objective."

        #
        # Count number of constraints and build the ampl_repns
        #
        n_ranges = 0
        n_single_sided_ineq = 0
        n_equals = 0
        n_nonlinear_constraints = 0
        ConVars = set()
        ConNonlinearVars = set()
        nnz_grad_constraints = 0
        for key, C in Con.iteritems():
            if C.trivial:
                continue
            for i, constraint_data in C._data.iteritems():
                ampl_repn = generate_ampl_repn(constraint_data.body)
                constraint_data.ampl_repn = ampl_repn
                if is_constant(ampl_repn):
                    continue

                # Merge variables from this expression
                all_vars = ampl_repn._linear_terms_var.copy()
                all_vars.update(ampl_repn._nonlinear_vars)
                Vars.update(all_vars)
                nnz_grad_constraints += len(all_vars)
                ConVars.update(all_vars.keys())
                ConNonlinearVars.update(ampl_repn._nonlinear_vars.keys())
                if ampl_repn.is_nonlinear():
                    n_nonlinear_constraints += 1

                L = None
                U = None
                if constraint_data.lower is not None: L = self._get_bound(constraint_data.lower)
                if constraint_data.upper is not None: U = self._get_bound(constraint_data.upper)
                if (L is None and U is None):
                    # constraint with NO bounds ???
                    raise ValueError, "Constraint " + str(con) +\
                                         "[" + str(i) + "]: lower bound = None and upper bound = None"
                elif (L is None or U is None):
                    # one sided inequality
                    n_single_sided_ineq += 1
                elif (L > U):
                    msg = 'Constraint %s[%s]: lower bound greater than upper' \
                          ' bound (%s > %s)'
                    raise ValueError, msg % (str(key), str(i), str(L), str(U) )
                elif (L == U):
                    # equality
                    n_equals += 1
                else:
                    # double sided inequality
                    # both are not none and they are valid
                    n_ranges += 1
        
        Ilist = []
        Blist = []
        Vlist = []
        keys = Vars.keys()
        keys.sort()
        for ndx in keys:
            var = Vars[ndx].var
            if isinstance(var.domain, IntegerSet):
                Ilist.append(ndx)
            elif isinstance(var.domain, BooleanSet):
                L = Vars[ndx].lb
                U = Vars[ndx].ub
                if L is None or U is None:
                    raise ValueError, "Variable " + str(var.label) +\
                          "is binary, but does not have lb and ub set"
                elif value(L) == 0 and value(U) == 1:
                    Blist.append(ndx)
                else:
                    Ilist.append(ndx)
            else:
                Vlist.append(ndx)

        Vlist.extend(Blist)
        Vlist.extend(Ilist)

        # put the ampl variable id into the variable
        var_id_ctr=0
        for var_name in Vlist:
            Vars[var_name].ampl_var_id = var_id_ctr
            var_id_ctr += 1

        #
        # Print Header
        #
        # LINE 1
        #
        print >>OUTPUT,"g3 1 1 0\t# problem",model.name
        #
        # LINE 2
        #
        print >>OUTPUT, "", len(Vars), str(n_single_sided_ineq+n_ranges+n_equals), str(n_objs), str(n_ranges), str(n_equals) +\
                "\t# vars, constraints, objectives, general inequalities, equalities"
        
        #
        # LINE 3
        #
        print >>OUTPUT, " " + str(n_nonlinear_constraints),  str(n_nonlinear_objs) + "\t# nonlinear constraints, objectives"
        #
        # LINE 4
        #
        print >>OUTPUT, " 0 0\t# network constraints: nonlinear, linear"
        #
        # LINE 5
        #
        Nonlinear_Vars_in_Objs_and_Constraints = ObjNonlinearVars.intersection(ConNonlinearVars)
        print >>OUTPUT, " " + str(len(ConNonlinearVars)), \
              str(len(ObjNonlinearVars)), str(len(Nonlinear_Vars_in_Objs_and_Constraints))\
              +"\t# nonlinear vars in constraints, objectives, both"
        #
        # LINE 6
        #
        print >>OUTPUT, " 0 0 0 1\t# linear network variables; functions; arith, flags"
        #
        # LINE 7
        #
        BIset = set(Blist).union(set(Ilist)) # maybe Blist and Ilist should be sets
        Discrete_Nonlinear_Vars_Both = Nonlinear_Vars_in_Objs_and_Constraints.intersection(BIset)
        Discrete_Nonlinear_Vars_Con_Only = ConNonlinearVars.intersection(BIset).difference(Discrete_Nonlinear_Vars_Both)
        Discrete_Nonlinear_Vars_Obj_Only = ObjNonlinearVars.intersection(BIset).difference(Discrete_Nonlinear_Vars_Both)
        n_int_nonlinear_b = len(Discrete_Nonlinear_Vars_Both)
        n_int_nonlinear_c = len(Discrete_Nonlinear_Vars_Con_Only)
        n_int_nonlinear_o = len(Discrete_Nonlinear_Vars_Obj_Only)
        print >>OUTPUT, " " + str(len(Blist)), str(len(Ilist)), n_int_nonlinear_b, n_int_nonlinear_c,\
                n_int_nonlinear_o, "\t# discrete variables: binary, integer, nonlinear (b,c,o)"
        #
        # LINE 8
        #
        # objective info computed above
        print >>OUTPUT, " " + str(nnz_grad_constraints), str(len(ObjVars)) + \
            "\t# nonzeros in Jacobian, obj. gradient"
        #
        # LINE 9
        #
        print >>OUTPUT, " 0 0\t# max name lengths: constraints, variables"
        #
        # LINE 10
        #
        print >>OUTPUT, " 0 0 0 0 0\t# common exprs: b,c,o,c1,o1"
        #
        # "C" lines
        #
        nc = 0
        for key, C in Con.iteritems():
            if C.trivial:
                continue
            for ndx, constraint_data in C._data.iteritems():
                ampl_repn = constraint_data.ampl_repn
                if is_constant(ampl_repn):
                    continue
                print >>OUTPUT, "C" + str(nc) + "\t#" + str(key) + "[" + str(ndx) + "]"
                nc += 1
                
                if ampl_repn.is_linear():
                    print >>OUTPUT, "n0"
                else:
                    assert(not ampl_repn._nonlinear_expr is None)
                    self._print_nonlinear_terms_NL(OUTPUT, ampl_repn._nonlinear_expr)
                                    
        #
        # "O" lines
        #
        no = 0
        for key in Obj:
            k = 0
            if Obj[key].sense == maximize:
                k = 1
            for ndx in Obj[key]:
                ampl_repn = Obj[key][ndx].ampl_repn
                if ampl_repn.is_constant():
                    pass
#                    continue
                print >>OUTPUT, "O" + str(no) + " "+str(k)+"\t#" + str(key) + "[" + str(ndx) + "]"
                no += 1

                if ampl_repn.is_linear():
                    print >>OUTPUT, "n" + str(ampl_repn._constant)
                else:
                    assert(not ampl_repn._nonlinear_expr is None)
                    if ampl_repn._constant != 0.0:
                        print >>OUTPUT, "o0", " #+"
                        print >>OUTPUT, "n" + str(ampl_repn._constant)
                    
                    self._print_nonlinear_terms_NL(OUTPUT, ampl_repn._nonlinear_expr)

        #
        # "x" lines
        #
        # variable initialization
        n_var_inits = 0
        for (var_name, var) in Vars.iteritems():
            if var.initial != None:
                n_var_inits += 1

        print >>OUTPUT, "x" + str(n_var_inits)
        for var_name in Vlist:
            var = Vars[var_name]
            if var.initial != None:
                print >>OUTPUT, str(var.ampl_var_id) + " " + str(var.initial) + " # " + var_name + " initial"
        
        #
        # "r" lines
        #
        print >>OUTPUT, "r"
        nc=0
        for key, C in Con.iteritems():

            if C.trivial:
                continue

            for ndx, constraint_data in C._data.iteritems():
                ampl_repn = constraint_data.ampl_repn
                if ampl_repn.is_constant():
                    continue

                offset = ampl_repn._constant

                if constraint_data._equality:
                    print >>OUTPUT, "4", self._get_bound(constraint_data.lower)-offset,
                else:
                    if constraint_data.lower is None:
                        if constraint_data.upper is None:
                            print >>OUTPUT,"3",
                        else:
                            print >>OUTPUT,"1", self._get_bound(constraint_data.upper)-offset,
                    else:
                        if constraint_data.upper is None:
                            print >>OUTPUT,"2", self._get_bound(constraint_data.lower)-offset,
                        else:
                            print >>OUTPUT,"0",\
                                    self._get_bound(constraint_data.lower)-offset,\
                                    self._get_bound(constraint_data.upper)-offset,
                print >>OUTPUT, " # c"+ str(nc)+"  "+constraint_data.label #str(key) + "[" + str(ndx) + "]"
                symbol_map["c" + str(nc)] = constraint_data.label #str(key) + "[" + str(ndx) + "]"
                nc += 1
        #
        # "b" lines
        #
        print >>OUTPUT, "b"
        for ndx in Vlist:
            vv = Vars[ndx]
            var = vv.var
            if isinstance(var.domain, BooleanSet):
                print >>OUTPUT, "0 0 1",
                print >>OUTPUT, " # v"+str(vv.ampl_var_id)+"  "+vv.label
                symbol_map["v"+str(vv.ampl_var_id)] = vv.label
                continue
            L = vv.lb
            U = vv.ub
            if L is not None:
                Lv = str(value(L))
                if U is not None:
                    Uv = str(value(U))
                    if Lv == Uv:
                        print >>OUTPUT, "4" , Lv,
                    else:
                        print >>OUTPUT, "0", Lv, Uv,
                else:
                    print >>OUTPUT, "2", Lv,
            elif U is not None:
                print >>OUTPUT, "1", str(value(U)),
            else:
                print >>OUTPUT, "3",
            print >>OUTPUT, " # v"+str(vv.ampl_var_id)+"  "+vv.label
            symbol_map["v"+str(vv.ampl_var_id)] = vv.label
        #
        # "k" lines
        #

        ktot = 0
        cu = {}
        for i in xrange(len(Vars)):
            cu[i] = 0

        for key, C in Con.iteritems():
            if C.trivial:
                continue
            for cndx, constraint_data in C._data.iteritems():
                con_vars = set(constraint_data.ampl_repn._linear_terms_var.keys())
                con_vars.update(set(constraint_data.ampl_repn._nonlinear_vars.keys()))
                for d in con_vars:
                    cu[Vars[d].ampl_var_id] += 1
        
        n1 = len(Vars) - 1
        print >>OUTPUT, "k" + str(n1)
        ktot = 0
        for i in xrange(n1):
                ktot += cu[i]
                print >>OUTPUT, ktot

        #
        # "J" lines
        #
        nc = 0
        for key, con in Con.iteritems():
            if con.trivial:
                continue
            for ndx, constraint_data in con._data.iteritems():
                ampl_repn = constraint_data.ampl_repn
                # we should probably have this set already created in ampl_repn
                con_vars = set(ampl_repn._linear_terms_var.keys())
                con_vars.update(set(ampl_repn._nonlinear_vars.keys()))
                print >>OUTPUT, "J" + str(nc), len(con_vars), " # ", constraint_data.label
                nc += 1
                grad_entries = {}
                for con_var in con_vars:
                    if con_var in ampl_repn._linear_terms_var:
                        grad_entries[Vars[con_var].ampl_var_id] = ampl_repn._linear_terms_coef[con_var]
                    else:
                        assert(con_var in ampl_repn._nonlinear_vars)
                        grad_entries[Vars[con_var].ampl_var_id] = 0

                sorted_keys = grad_entries.keys()
                sorted_keys.sort()
                for var_id in sorted_keys:
                    print >>OUTPUT, var_id, " ", grad_entries[var_id] 
        #
        # "G" lines
        #
        no = 0
        grad_entries = {}
        for key in Obj:
            if Obj[key].trivial:
                continue
            obj = Obj[key]
            for ndx in obj:
                ampl_repn = obj[ndx].ampl_repn
                obj_vars = set(ampl_repn._linear_terms_var.keys())
                obj_vars.update(set(ampl_repn._nonlinear_vars.keys()))
                if len(obj_vars) == 0:
                    pass
                else:
                    print >>OUTPUT, "G" + str(no), len(obj_vars)
                no += 1

                for obj_var in obj_vars:
                    if obj_var in ampl_repn._linear_terms_var:
                        grad_entries[Vars[obj_var].ampl_var_id] = ampl_repn._linear_terms_coef[obj_var]
                    else:
                        assert(obj_var in ampl_repn._nonlinear_vars)
                        grad_entries[Vars[obj_var].ampl_var_id] = 0

        sorted_keys = grad_entries.keys()
        sorted_keys.sort()
        for var_id in sorted_keys:
            print >>OUTPUT, var_id, " ", grad_entries[var_id]

        return symbol_map
    
class ampl_representation:
    def __init__(self):
        self._constant = 0
        self._linear_terms_var = {}
        self._linear_terms_coef = {}
        self._nonlinear_expr = None
        self._nonlinear_vars = {}

    def clone(self):
        clone = ampl_representation()
        clone._constant = self.constant
        clone._linear_terms_var = self._linear_terms_var
        clone._linear_terms_coef = self._linear_terms_coef
        clone._nonlinear_expr = self._nonlinear_expr
        clone._nonlinear_vars = self._nonlinear_vars
        return clone
        
    def is_constant(self):
        if len(self._linear_terms_var) == 0 and self._nonlinear_expr is None:
#            assert(len(self._linear_terms_coef) == 0)
#            assert(len(self._nonlinear_vars) == 0)
            return True
        return False
    
    def is_linear(self):
        if self._nonlinear_expr is None:
            assert(len(self._nonlinear_vars) == 0)
            return True

    def is_nonlinear(self):
        if not self._nonlinear_expr is None:
            return True
        return False

    def needs_sum(self):
        num_components = 0
        if self._constant != 0:
            num_components += 1
        
        num_components += len(self._linear_terms_var)
        
        if not self._nonlinear_expr is None:
            num_components += 1
            
        if num_components > 1:
            return True
        return False

    def create_expr(self):
        if self.needs_sum() == True:
            ret_expr = expr._SumExpression()
            if self.is_constant():
                ret_expr._const = self._constant
            
            for (var_name, var) in self._linear_terms_var.iteritems():
                ret_expr.add(self._linear_terms_coef[var_name], var)

            if self.is_nonlinear():
                ret_expr.add(1.0, self._nonlinear_expr)

            return ret_expr
        else:            
            # we know there is only one component present in ampl_repn
            if self.is_constant():
                return as_numeric(self._constant)
            elif self.is_nonlinear():
                return self._nonlinear_expr
            else:
                assert( len(self._linear_terms_var) == 1 )
                return self._linear_terms_var[self._linear_terms_var.keys()[0]]
            
            
        
def generate_ampl_repn(exp):
    ampl_repn = ampl_representation()
#    exp.pprint()
    #
    # recurse to the bottom of the identity expression
    #
    while isinstance(exp, expr._IdentityExpression):
        exp = exp._args[0]

    #
    # Expression
    #
    if isinstance(exp,expr.Expression):
        #
        # Sum
        #
        if isinstance(exp,expr._SumExpression):
            ampl_repn._constant = float(exp._const)
            ampl_repn._nonlinear_expr = None
            for i in xrange(len(exp._args)):
                exp_coef = exp._coef[i]
                child_repn = generate_ampl_repn(exp._args[i])
                # adjust the constant
                ampl_repn._constant += exp_coef*child_repn._constant

                # adjust the linear terms
                for (var_name,var) in child_repn._linear_terms_var.iteritems():
                    if var_name in ampl_repn._linear_terms_var:
                        ampl_repn._linear_terms_coef[var_name] += exp_coef*child_repn._linear_terms_coef[var_name]
                    else:
                        ampl_repn._linear_terms_var[var_name] = var
                        ampl_repn._linear_terms_coef[var_name] = exp_coef*child_repn._linear_terms_coef[var_name]
            
                # adjust the nonlinear terms
                if not child_repn._nonlinear_expr is None:
                    if ampl_repn._nonlinear_expr is None:
#                        ampl_repn._nonlinear_expr = expr._SumExpression([child_repn._nonlinear_expr],[exp_coef])
                        ampl_repn._nonlinear_expr = [(exp_coef, child_repn._nonlinear_expr)]
                    else:
#                        assert(isinstance(ampl_repn._nonlinear_expr, expr._SumExpression))
                        assert(len(ampl_repn._nonlinear_expr) > 0)
                        ampl_repn._nonlinear_expr.append((exp_coef, child_repn._nonlinear_expr))

                # adjust the nonlinear vars
                ampl_repn._nonlinear_vars.update(child_repn._nonlinear_vars)
                
            return ampl_repn
   
        #
        # Product
        #
        elif isinstance(exp,expr._ProductExpression):
            #
            # Iterate through the denominator.  If they aren't all constants, then
            # simply return this expresion.
            #
            denom=1.0
            for e in exp._denominator:
                if e.fixed_value():
                    #print 'Z',type(e),e.fixed
                    denom *= e.value
                elif e.is_constant():
                    denom *= e()
                else:
                    ampl_repn._nonlinear_expr = exp
                    break
                if denom == 0.0:
                    print "Divide-by-zero error - offending sub-expression:"
                    e.pprint()
                    raise ZeroDivisionError

            if not ampl_repn._nonlinear_expr is None:
                # we have a nonlinear expression ... build up all the vars
                for e in exp._denominator:
                    arg_repn = generate_ampl_repn(e)
                    ampl_repn._nonlinear_vars.update(arg_repn._linear_terms_var)
                    ampl_repn._nonlinear_vars.update(arg_repn._nonlinear_vars)

                for e in exp._numerator:
                    arg_repn = generate_ampl_repn(e)
                    ampl_repn._nonlinear_vars.update(arg_repn._linear_terms_var)
                    ampl_repn._nonlinear_vars.update(arg_repn._nonlinear_vars)
                return ampl_repn

            #
            # OK, the denominator is a constant.
            #
            # build up the ampl_repns for the numerator
            n_linear_args = 0
            n_nonlinear_args = 0
            arg_repns = list()
            for i in xrange(len(exp._numerator)):
                e = exp._numerator[i]
                e_repn = generate_ampl_repn(e)
                arg_repns.append(e_repn)
                if e_repn.is_nonlinear():
                    n_nonlinear_args += 1
                elif not e_repn.is_constant():
                    n_linear_args += 1

            is_nonlinear = False
            if n_linear_args > 1 or n_nonlinear_args > 0:
                is_nonlinear = True
                
                
            if is_nonlinear == True:
                # do like AMPL and simply return the expression
                # without extracting the potentially linear part
                ampl_repn = ampl_representation()
                ampl_repn._nonlinear_expr = exp
                for repn in arg_repns:
                    ampl_repn._nonlinear_vars.update(repn._linear_terms_var)
                    ampl_repn._nonlinear_vars.update(repn._nonlinear_vars)
                return ampl_repn

            else: # is linear or constant
                ampl_repn = current_repn = arg_repns[0]
                for i in xrange(1,len(arg_repns)):
                    e_repn = arg_repns[i]
                    ampl_repn = ampl_representation()

                    # const_c * const_e
                    ampl_repn._constant = current_repn._constant * e_repn._constant
    
                    # const_e * L_c
                    if e_repn._constant != 0.0:
                        for (var_name, var) in current_repn._linear_terms_var.iteritems():
                            ampl_repn._linear_terms_coef[var_name] = current_repn._linear_terms_coef[var_name] * e_repn._constant 
                            ampl_repn._linear_terms_var[var_name] = var
    
                    # const_c * L_e
                    if current_repn._constant != 0.0:
                        for (e_var_name,e_var) in e_repn._linear_terms_var.iteritems():
                            if e_var_name in ampl_repn._linear_terms_var:
                                ampl_repn._linear_terms_coef[e_var_name] += current_repn._constant * e_repn._linear_terms_coef[e_var_name]
                            else:
                                ampl_repn._linear_terms_coef[e_var_name] = current_repn._constant * e_repn._linear_terms_coef[e_var_name] 
                                ampl_repn._linear_terms_var[e_var_name] = e_var
                
                current_repn = ampl_repn

            # now deal with the product expression's coefficient that needs
            # to be applied to all parts of the ampl_repn
            ampl_repn._constant *= exp.coef/denom
            for var_name in ampl_repn._linear_terms_coef:
                ampl_repn._linear_terms_coef[var_name] *= exp.coef/denom
            assert(ampl_repn._nonlinear_expr is None)
            assert(len(ampl_repn._nonlinear_vars) == 0)

            return ampl_repn

        elif isinstance(exp,expr._PowExpression):
            assert(len(exp._args) == 2)
            base_repn = generate_ampl_repn(exp._args[0])
            exponent_repn = generate_ampl_repn(exp._args[1])

            if base_repn.is_constant() and exponent_repn.is_constant():
                ampl_repn._constant = base_repn._constant**exponent_repn._constant
            elif exponent_repn.is_constant() and exponent_repn._constant == 1.0:
                # not sure if this is necessary (may be able to just return base_repn
                ampl_repn = base_repn.clone()
            elif exponent_repn.is_constant() and exponent_repn._constant == 0.0:
                ampl_repn._constant = 1.0
            else:
                # this might be overkill, creating another sum expression
#                base_expr = base_repn.create_expr()
#                exponent_expr = exponent_repn.create_expr()
#                ampl_repn._nonlinear_expr = expr._PowExpression(base_expr, exponent_expr)

                # instead, let's just return the expression we are given and only
                # use the ampl_repn for the vars
                ampl_repn._nonlinear_expr = exp
                ampl_repn._nonlinear_vars = base_repn._nonlinear_vars
                ampl_repn._nonlinear_vars.update(exponent_repn._nonlinear_vars)
                ampl_repn._nonlinear_vars.update(base_repn._linear_terms_var)
                ampl_repn._nonlinear_vars.update(exponent_repn._linear_terms_var)

            return ampl_repn
              
        elif isinstance(exp,expr._IntrinsicFunctionExpression):
            assert(len(exp._args) == 1)

            # this is inefficient since it is using much more than what we need
            child_repn = generate_ampl_repn(exp._args[0])
#            assert(child_repn.is_constant() == False)
            
            ampl_repn._nonlinear_expr = exp
            ampl_repn._nonlinear_vars = child_repn._nonlinear_vars
            ampl_repn._nonlinear_vars.update(child_repn._linear_terms_var)
            return ampl_repn
        #
        # ERROR
        #
        else:
            raise ValueError, "Unsupported expression type: "+str(exp)
    #
    # Constant
    #
    elif exp.fixed_value():
        ampl_repn._constant = float(exp.value)
        return ampl_repn
    #
    # Variable
    #
    elif type(exp) is _VarValue or exp.type() is Var:
        ampl_repn._linear_terms_coef = {exp.label: 1.0}
        ampl_repn._linear_terms_var = {exp.label: exp}
        return ampl_repn
    #
    # ERROR
    #
    else:
       raise ValueError, "Unexpected expression type: "+str(exp)

